# -*- coding: utf-8 -*-
"""ThreatConnect Case Management"""
from .assignee import Assignee, User, Users
from .artifact import Artifact, Artifacts
from .artifact_type import ArtifactType, ArtifactTypes
from .case import Case, Cases
from .note import Note, Notes
from .tag import Tag, Tags
from .task import Task, Tasks
from .workflow_event import WorkflowEvent, WorkflowEvents
from .workflow_template import WorkflowTemplate, WorkflowTemplates


class CaseManagement:
    """Case Management Class for TcEx

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
    """

    def __init__(self, tcex):
        """Initialize Class properties."""
        self.tcex = tcex

    def artifact(self, **kwargs):
        """Return a instance of Artifact object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            analytics_priority (str, kwargs): [Read-Only] The **Analytics Priority**
                for the Artifact.
            analytics_priority_level (int, kwargs): [Read-Only] The **Analytics Priority Level** for
                the Artifact.
            analytics_score (int, kwargs): [Read-Only] The **Analytics Score** for the Artifact.
            analytics_type (str, kwargs): [Read-Only] The **Analytics Type** for the Artifact.
            artifact_type (ArtifactType, kwargs): [Read-Only] The **Artifact Type**
                for the Artifact.
            case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Artifact.
            case_xid (str, kwargs): [Required (alt: caseId)] The **Case Xid** for the Artifact.
            date_added (str, kwargs): [Read-Only] The **Date Added** for the Artifact.
            file_data (str, kwargs): Base64 encoded file attachment required only for certain
                artifact types
            intel_type (str, kwargs): The **Intel Type** for the Artifact.
            links (Link, kwargs): The **Links** for the Artifact.
            notes (Note, kwargs): a list of Notes corresponding to the Artifact
            parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Artifact.
            source (str, kwargs): The **Source** for the Artifact.
            summary (str, kwargs): [Required] The **Summary** for the Artifact.
            task (case_management.task.Task, kwargs): [Read-Only] The **Task** for the Artifact.
            task_id (int, kwargs): the ID of the task which the Artifact references
            task_xid (str, kwargs): the XID of the task which the Artifact references
            type (str, kwargs): [Required] The **Type** for the Artifact.
        """
        return Artifact(self.tcex, **kwargs)

    def artifacts(self, **kwargs):
        """Return a instance of Artifacts object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Artifact. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Artifacts objects.
        """
        return Artifacts(self.tcex, **kwargs)

    def artifact_type(self, **kwargs):
        """Return a instance of Artifact Type object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            data_type (str, kwargs): [Read-Only] The **Data Type** for the Artifact Type.
            description (str, kwargs): [Read-Only] The **Description** for the Artifact Type.
            intel_type (str, kwargs): [Read-Only] The **Intel Type** for the Artifact Type.
            name (str, kwargs): [Read-Only] The **Name** for the Artifact Type.
        """
        return ArtifactType(self.tcex, **kwargs)

    def artifact_types(self, **kwargs):
        """Return a instance of Artifact Types object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Artifact. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent
                while retrieving the Artifact Type objects.
        """
        return ArtifactTypes(self.tcex, **kwargs)

    @staticmethod
    def assignee(**kwargs):
        """Return a instance of Assignee object.

        For User type the user_name or id fields is required and
        for the Group type the name or id is required.

        Args:
            first_name (str, kwargs): The first name of the User.
            id (id, kwargs): The id of the User.
            name (str, kwargs): The name of the Group.
            last_name (str, kwargs): The last name of the user.
            pseudonym (str, kwargs): The pseudonym of the User.
            role (str, kwargs): The role of the User.
            user_name (str, kwargs): The user name of the User.
            type (str, kwargs): The assignee type. Default to User.
        """
        return Assignee(**kwargs)

    def case(self, **kwargs):
        """Return a instance of Case object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            artifacts (Artifact, kwargs): a list of Artifacts corresponding to the Case
            assignee (Assignee, kwargs): the user or group Assignee object for the Case
            created_by (User, kwargs): [Read-Only] The **Created By** for the Case.
            date_added (str, kwargs): [Read-Only] The **Date Added** for the Case.
            description (str, kwargs): The **Description** for the Case.
            name (str, kwargs): [Required] The **Name** for the Case.
            notes (Note, kwargs): a list of Notes corresponding to the Case
            owner (str, kwargs): The Org/Owner name for the Case.
            related (Case, kwargs): The **Related** for the Case.
            resolution (str, kwargs): The **Resolution** for the Case.
            severity (str, kwargs): [Required] The **Severity** for the Case.
            status (str, kwargs): [Required] The **Status** for the Case.
            tags (tcex.case_management.tag.Tag, kwargs): a list of Tags corresponding to the Case
                (NOTE: Setting this parameter will replace any existing tag(s) with
                the one(s) specified)
            tasks (case_management.task.Task, kwargs): a list of Tasks corresponding to the Case
            user_access (User, kwargs): a list of Users that, when defined, are the only
                ones allowed to view or edit the Case
            workflow_events (WorkflowEvent, kwargs): The **Events** for the Case.
            workflow_template (WorkflowTemplate, kwargs): the Template that the Case is
                populated by.
            xid (str, kwargs): The **Xid** for the Case.
        """
        return Case(self.tcex, **kwargs)

    def cases(self, **kwargs):
        """Return a instance of Cases object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Artifact. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Case objects.
        """
        return Cases(self.tcex, **kwargs)

    def create_entity(self, entity, owner):
        """Create a CM object provided a dict and owner."""
        entity_type = entity.get('type').lower()
        entity_type = entity_type.replace(' ', '_')
        try:
            obj = getattr(self, entity_type)(**entity)
        except AttributeError:
            return None

        r = obj.submit()
        data = {'status_code': r.status_code}
        if r.ok:
            data.update(r.json().get('data', {}))
            data['main_type'] = 'Case_Management'
            data['sub_type'] = entity_type
            data['owner'] = owner

        return data

    def note(self, **kwargs):
        """Return a instance of Note object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            artifact (Artifact, kwargs): [Read-Only] The **Artifact** for the Note.
            artifact_id (int, kwargs): the ID of the Artifact on which to apply the Note
            author (int, kwargs): [Read-Only] The **Author** for the Note.
            case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Note.
            case_xid (str, kwargs): [Required (alt: caseId)]
            date_added (str, kwargs): [Read-Only] The **Date Added** for the Note.
            edited (boolean, kwargs): [Read-Only] The **Edited** for the Note.
            last_modified (str, kwargs): [Read-Only] The **Last Modified** for the Note.
            parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Note.
            summary (str, kwargs): [Read-Only] The **Summary** for the Note.
            task (case_management.task.Task, kwargs): [Read-Only] The **Task** for the Note.
            task_id (int, kwargs): the ID of the Task on which to apply the Note
            task_xid (str, kwargs): the XID of the Task on which to apply the Note
            text (str, kwargs): [Required] The **Text** for the Note.
            workflow_event (WorkflowEvent, kwargs): [Read-Only] The **Workflow Event** for the Note.
            workflow_event_id (int, kwargs): the ID of the Event on which to apply the Note
        """
        return Note(self.tcex, **kwargs)

    def notes(self, **kwargs):
        """Return a instance of Notes object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Note. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Notes objects.
        """
        return Notes(self.tcex, **kwargs)

    def tag(self, **kwargs):
        """Return a instance of Tag object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            cases (Case, kwargs): [Read-Only] The **Cases** for the Tag.
            description (str, kwargs): a brief description of the Tag
            last_used (str, kwargs): [Read-Only] The **Last Used** for the Tag.
            name (str, kwargs): [Required] The **Name** for the Tag.
            owner (str, kwargs): [Read-Only] The **Owner** for the Tag.
        """
        return Tag(self.tcex, **kwargs)

    def tags(self, **kwargs):
        """Return a instance of Tags object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Tag. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Tag objects.
        """
        return Tags(self.tcex, **kwargs)

    def task(self, **kwargs):
        """Return a instance of Task object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            artifacts (Artifact, kwargs): a list of Artifacts corresponding to the Task
            assignee (Assignee, kwargs): the user or group Assignee object for the Task
            case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Task.
            case_xid (str, kwargs): [Required (alt: caseId)] The **Case Xid** for the Task.
            completed_by (str, kwargs): [Read-Only] The **Completed By** for the Task.
            completed_date (str, kwargs): the completion date of the Task
            config_playbook (str, kwargs): [Read-Only] The **Config Playbook** for the Task.
            config_task (dict, kwargs): [Read-Only] The **Config Task** for the Task.
            dependent_on_id (str, kwargs): [Read-Only] The **Dependent On ID** for the Task.
            description (str, kwargs): The **Description** for the Task.
            due_date (str, kwargs): the due date of the Task
            duration (int, kwargs): [Read-Only] The **Duration** for the Task.
            id_dependent_on (int, kwargs): [Read-Only] The **Id_Dependent On** for the Task.
            name (str, kwargs): [Required] The **Name** for the Task.
            notes (Note, kwargs): a list of Notes corresponding to the Task
            parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Task.
            required (bool, kwargs): [Read-Only] The **Required** flag for the Task.
            status (str, kwargs): The **Status** for the Task.
            workflow_phase (int, kwargs): the phase of the workflow
            workflow_step (int, kwargs): the step of the workflow
            xid (str, kwargs): The **Xid** for the Task.
        """
        return Task(self.tcex, **kwargs)

    def tasks(self, **kwargs):
        """Return a instance of Tasks object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Task. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Task objects.
        """
        return Tasks(self.tcex, **kwargs)

    @staticmethod
    def user(**kwargs):
        """Return a instance of User object.

        Args:
            first_name (str, kwargs): The first name of the user.
            id (id, kwargs): The id of the user.
            last_name (str, kwargs): The last name of the user.
            pseudonym (str, kwargs): The pseudonym of the user.
            role (str, kwargs): The role of the user.
            user_name (str, kwargs): The user name of the user.
        """
        return User(**kwargs)

    @staticmethod
    def users(users):
        """Sub class of the Cases object. Used to map the users to.

        Args:
            users (list): A array of user data
        """
        return Users(users)

    def workflow_event(self, **kwargs):
        """Return a instance of Workflow Event object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            case_id (int, kwargs): [Required (alt: caseXid)] The **Case Id** for the Workflow Event.
            case_xid (str, kwargs): [Required (alt: caseId)] The **Case Xid**
                for the Workflow Event.
            date_added (str, kwargs): [Read-Only] The **Date Added** for the Workflow Event.
            deleted (bool, kwargs): [Read-Only] The **Deleted** flag for the Workflow Event.
            deleted_reason (str, kwargs): the reason for deleting the event (required
                input for DELETE operation only)
            event_date (str, kwargs): the time that the Event is logged
            link (str, kwargs): [Read-Only] The **Link** for the Workflow Event.
            link_text (str, kwargs): [Read-Only] The **Link Text** for the Workflow Event.
            notes (Note, kwargs): a list of Notes corresponding to the Event
            parent_case (Case, kwargs): [Read-Only] The **Parent Case** for the Workflow Event.
            summary (str, kwargs): [Required] The **Summary** for the Workflow Event.
            system_generated (bool, kwargs): [Read-Only] The **System Generated**
                flag for the Workflow Event.
            user (User, kwargs): [Read-Only] The **User** for the Workflow Event.
        """
        return WorkflowEvent(self.tcex, **kwargs)

    def workflow_events(self, **kwargs):
        """Return a instance of Workflow Events object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Artifact. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Workflow Event objects.
        """
        return WorkflowEvents(self.tcex, **kwargs)

    def workflow_template(self, **kwargs):
        """Return a instance of Workflow Template object.

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            active (bool, kwargs): [Read-Only] The **Active** flag for the Workflow Template.
            assignee (Assignee, kwargs): [Read-Only] The **Assignee** for the Workflow Template.
            cases (Case, kwargs): [Read-Only] The **Cases** for the Workflow Template.
            config_artifact (str, kwargs): [Read-Only] The **Config Artifact** for the Workflow
                Template.
            config_playbook (str, kwargs): [Read-Only] The **Config Playbook** for the Workflow
                Template.
            config_task (dict, kwargs): [Read-Only] The **Config Task** for the Workflow Template.
            description (str, kwargs): The **Description** for the Workflow Template.
            name (str, kwargs): [Required] The **Name** for the Workflow Template.
            target_type (str, kwargs): [Read-Only] The **Target Type** for the Workflow Template.
            version (int, kwargs): The **Version** for the Workflow Template.
        """
        return WorkflowTemplate(self.tcex, **kwargs)

    def workflow_templates(self, **kwargs):
        """Return a instance of Workflow Templates object.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            # Example of params input
            {
                'result_limit': 100,  # How many results are retrieved.
                'result_start': 10,  # Starting point on retrieved results.
                'fields': ['caseId', 'summary']  # Additional fields returned on the results
            }

        Args:
            tcex (TcEx): An instantiated instance of TcEx object.
            initial_response (dict, optional): Initial data in
                Case Object for Workflow Template. Defaults to None.
            tql_filters (list, optional): List of TQL filters. Defaults to None.
            params(dict, optional): Dict of the params to be sent while
                retrieving the Workflow Event objects.
        """
        return WorkflowTemplates(self.tcex, **kwargs)



# -*- coding: utf-8 -*-
"""TcEx Framework Logger module"""
import logging
import os
import pathlib
import platform
import sys

from .api_handler import ApiHandler, ApiHandlerFormatter
from .cache_handler import CacheHandler
from .rotating_file_handler_custom import RotatingFileHandlerCustom
from .thread_file_handler import ThreadFileHandler
from .trace_logger import TraceLogger


class Logger:
    """Framework logger module."""

    def __init__(self, tcex, logger_name):
        """Initialize Class Properties.

        Args:
            tcex (tcex.TcEx): Instance of TcEx class.
            logger_name (str): The name of the logger.
        """
        self.tcex = tcex
        self.logger_name = logger_name

    @property
    def _logger(self):
        """Return the logger. The default_args property is not available in init."""
        logging.setLoggerClass(TraceLogger)
        logger = logging.getLogger(self.logger_name)
        logger.setLevel(logging.TRACE)
        return logger

    @property
    def _formatter(self):
        """Return log formatter."""
        tx_format = (
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s '
            '(%(filename)s:%(funcName)s:%(lineno)d)'
        )
        return logging.Formatter(tx_format)

    @property
    def _formatter_thread_name(self):
        """Return log formatter."""
        tx_format = (
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s '
            '(%(filename)s:%(funcName)s:%(lineno)d:%(threadName)s)'
        )
        return logging.Formatter(tx_format)

    @property
    def log(self):
        """Return logger."""
        return self._logger

    @staticmethod
    def log_level(level):
        """Return proper level from string.

        Args:
            level (str): The logging level. Default to 'debug'.
        """
        level = level or 'debug'
        return logging.getLevelName(level.upper())

    def remove_handler_by_name(self, handler_name):
        """Remove a file handler by name.

        Args:
            handler_name (str): The handler name to remove.
        """
        for h in self._logger.handlers:
            if h.get_name() == handler_name:
                self._logger.removeHandler(h)
                break

    def replay_cached_events(self, handler_name='cache'):
        """Replay cached log events and remove handler."""
        for h in self._logger.handlers:
            if h.get_name() == handler_name:
                events = h.events
                self._logger.removeHandler(h)
                for event in events:
                    self._logger.handle(event)
                break

    def shutdown(self):
        """Close all handlers.

        Args:
            handler_name (str): The handler name to remove.
        """
        for h in self._logger.handlers:
            self._logger.removeHandler(h)

    def update_handler_level(self, level):
        """Update all handlers log level.

        Args:
            level (int, optional): The logging level. Defaults to None.
        """
        level = self.log_level(level)

        # update all handler logging levels
        for h in self._logger.handlers:
            h.setLevel(level)

    #
    # handlers
    #

    def add_api_handler(self, name='api', level=None):
        """Add API logging handler.

        Args:
            name (str, optional): The name of the handler. Defaults to 'api'.
            level (str, optional): The level value as a string. Defaults to None.
        """
        self.remove_handler_by_name(name)
        api = ApiHandler(self.tcex.session)
        api.set_name(name)
        api.setLevel(self.log_level(level))
        api.setFormatter(ApiHandlerFormatter())
        self._logger.addHandler(api)

    def add_cache_handler(self, name):
        """Add cache logging handler.

        Args:
            name (str): The name of the handler.
        """
        self.remove_handler_by_name(name)
        cache = CacheHandler()
        cache.set_name(name)
        # set logging level to INFO as event will typically be only those that happen before args
        # are processed
        cache.setLevel(self.log_level('trace'))
        cache.setFormatter(self._formatter)
        self._logger.addHandler(cache)

    def add_rotating_file_handler(
        self, name, filename, path, backup_count, max_bytes, level, formatter=None, mode='a'
    ):
        """Add a rotating file handler

        Args:
            name (str, optional): The name of the handler. Defaults to 'rfh'.
            filename (str): The name of the logfile.
            path (str): The path for the logfile.
            backup_count (int, optional): The maximum # of backup files. Defaults to 0.
            max_bytes (int): The max file size before rotating. Defaults to 0.
            level (str): The logging level. Defaults to None.
            formatter (str, optional): The logging formatter to use. Defaults to None.
            mode (str, optional): The write mode for the file. Defaults to 'a'.
        """
        self.remove_handler_by_name(name)
        formatter = formatter or self._formatter_thread_name

        # create customized handler
        fh = RotatingFileHandlerCustom(
            os.path.join(path, filename), backupCount=backup_count, maxBytes=max_bytes, mode=mode
        )
        fh.set_name(name)
        fh.setFormatter(formatter)
        fh.setLevel(self.log_level(level))
        self._logger.addHandler(fh)

    def add_stream_handler(self, name='sh', formatter=None, level=None):
        """Return stream logging handler.

        Args:
            name (str, optional): The name of the handler. Defaults to 'sh'.
            formatter (str, optional): The logging formatter to use. Defaults to None.
            level (int, optional): The logging level. Defaults to None.
        """
        self.remove_handler_by_name(name)
        formatter = formatter or self._formatter
        # create handler
        sh = logging.StreamHandler()
        sh.set_name(name)
        sh.setFormatter(formatter)
        sh.setLevel(self.log_level(level))
        self._logger.addHandler(sh)

    def add_thread_file_handler(self, name, filename, level, path, formatter=None):
        """Add File logging handler.

        Args:
            name (str): The name of the handler.
            filename (str): The name of the logfile.
            level (int, optional): The logging level. Defaults to None.
            path (str): The path for the logfile.
            formatter (str, optional): The logging formatter to use. Defaults to None.
        """
        self.remove_handler_by_name(name)
        formatter = formatter or self._formatter
        # create customized handler
        fh = ThreadFileHandler(os.path.join(path, filename))
        fh.set_name(name)
        fh.setFormatter(formatter)
        fh.setLevel(self.log_level(level))
        self._logger.addHandler(fh)

    #
    # App info logging
    #

    def log_info(self, args):
        """Send System and App data to logs."""
        self._log_platform()
        self._log_app_data()
        self._log_python_version()
        self._log_tcex_version()
        self._log_tcex_path()
        self._log_tc_proxy(args)

    def _log_app_data(self):
        """Log the App data information."""
        # Best Effort
        try:
            self.log.info(f'App Name: {self.tcex.ij.display_name}')
            if self.tcex.ij.features:
                self.log.info(f"App Features: {','.join(self.tcex.ij.features)}")
            self.log.info(f'App Minimum ThreatConnect Version: {self.tcex.ij.min_server_version}')
            self.log.info(f'App Runtime Level: {self.tcex.ij.runtime_level}')
            self.log.info(f'App Version: {self.tcex.ij.program_version}')
            if self.tcex.ij.commit_hash is not None:
                self.log.info(f'App Commit Hash: {self.tcex.ij.commit_hash}')
        except Exception:  # pragma: no cover
            pass

    def _log_platform(self):
        """Log the current Platform."""
        self.log.info(f'Platform: {platform.platform()}')

    def _log_python_version(self):
        """Log the current Python version."""
        self.log.info(
            f'Python Version: {sys.version_info.major}.'
            f'{sys.version_info.minor}.'
            f'{sys.version_info.micro}'
        )

    def _log_tc_proxy(self, args):
        """Log the proxy settings."""
        if args.tc_proxy_tc:
            self.log.info(f'Proxy Server (TC): {args.tc_proxy_host}:{args.tc_proxy_port}.')

    def _log_tcex_version(self):
        """Log the current TcEx version number."""
        self.log.info(f'TcEx Version: {__import__(__name__).__version__}')

    def _log_tcex_path(self):
        """Log the current TcEx path."""
        app_path = str(pathlib.Path().parent.absolute())
        full_path = str(pathlib.Path(__file__).parent.absolute())
        tcex_path = os.path.dirname(full_path.replace(app_path, ''))
        self.log.info(f'TcEx Path: {tcex_path}')



# -*- coding: utf-8 -*-
"""TcEx Utilities Module"""
import ipaddress
import os
import random
import re
import string
import uuid
from typing import Any, List

import pyaes

from .date_utils import DatetimeUtils


class Utils:
    """TcEx framework Utils Class

    Args:
        temp_path (str, optional): The path to write temp files.
    """

    def __init__(self, temp_path=None):
        """Initialize the Class properties."""
        self.temp_path = temp_path or '/tmp'

        # properties
        self._camel_pattern = re.compile(r'(?<!^)(?=[A-Z])')
        self._inflect = None
        self.variable_match = re.compile(fr'^{self.variable_pattern}$')
        self.variable_parse = re.compile(self.variable_pattern)

    def camel_to_snake(self, camel_string):
        """Return snake case string from a camel case string.

        Args:
            camel_string (str): The camel case input string.

        Returns:
            str: The snake case representation of input string.
        """
        return self._camel_pattern.sub('_', camel_string).lower()

    def camel_to_space(self, camel_string):
        """Return space case string from a camel case string.

        Args:
            camel_string (str): The camel case input string.

        Returns:
            str: The space representation of input string.
        """
        return self._camel_pattern.sub(' ', camel_string).lower()

    @property
    def datetime(self):
        """Return an instance of DatetimeUtils."""
        return DatetimeUtils()

    @staticmethod
    def decrypt_aes_cbc(key, ciphertext, iv=None):
        """Return AES CBC decrypted string.

        Args:
            key (bytes): The encryption key.
            ciphertext (bytes): The ciphertext to decrypt.
            iv (bytes, optional): The CBC initial vector.

        Returns:
            bytes: The encoded string.
        """
        iv = iv or b'\0' * 16

        # ensure key is bytes
        if isinstance(key, str):
            key = key.encode()

        # ensure iv is bytes
        if isinstance(iv, str):
            iv = iv.encode()

        aes_cbc_decrypt = pyaes.Decrypter(pyaes.AESModeOfOperationCBC(key, iv=iv))
        decrypted = aes_cbc_decrypt.feed(ciphertext)
        decrypted += aes_cbc_decrypt.feed()
        return decrypted

    @staticmethod
    def encrypt_aes_cbc(key, plaintext, iv=None):
        """Return AES CBC encrypted string.

        Args:
            key (bytes): The encryption key.
            plaintext (str|bytes): The text to encrypt.
            iv (bytes, optional): The CBC initial vector.

        Returns:
            bytes: The encoded string.
        """
        iv = iv or b'\0' * 16

        # ensure key is bytes
        if isinstance(key, str):
            key = key.encode()

        # ensure plaintext is bytes
        if isinstance(plaintext, str):
            plaintext = plaintext.encode()

        # ensure iv is bytes
        if isinstance(iv, str):
            iv = iv.encode()

        aes_cbc_encrypt = pyaes.Encrypter(pyaes.AESModeOfOperationCBC(key, iv=iv))
        encrypted = aes_cbc_encrypt.feed(plaintext)
        encrypted += aes_cbc_encrypt.feed()
        return encrypted

    @staticmethod
    def flatten_list(lst: List[Any]) -> List[Any]:
        """Flatten a list

        Will work for lists of lists to arbitrary depth
        and for lists with a mix of lists and single values

        Args:
            lst (list): The list to flatten

        Returns:
            list: the flattened list
        """
        flat_list = []
        for sublist in lst:
            if isinstance(sublist, list):
                for item in Utils.flatten_list(sublist):
                    flat_list.append(item)
            else:
                flat_list.append(sublist)

        return flat_list

    @property
    def inflect(self):
        """Return instance of inflect."""
        if self._inflect is None:
            import inflect

            self._inflect = inflect.engine()
        return self._inflect

    @staticmethod
    def is_cidr(possible_cidr_range: str) -> bool:
        """Return whether the possible_cidr_range is a CIDR range."""
        try:
            ipaddress.ip_address(possible_cidr_range)
        except ValueError:
            try:
                ipaddress.ip_interface(possible_cidr_range)
            except Exception:
                return False
            else:
                return True
        else:
            return False

    @staticmethod
    def is_ip(possible_ip: str) -> bool:
        """Return whether the possible_ip is an IP address range."""
        try:
            ipaddress.ip_address(possible_ip)
        except ValueError:
            return False
        else:
            return True

    @staticmethod
    def printable_cred(cred, visible=1, mask_char=None):
        """Return a printable (masked) version of the provided credential.

        Args:
            cred (str): The cred to print.
            visible (str): The number of characters at the beginning and
                ending of the cred to not mask.
            mask_char (str, Default: "*"): The character to use in the mask.

        Returns:
            str: The reformatted token.
        """
        mask_char = mask_char or '*'
        if cred is not None and len(cred) >= visible * 2:
            cred = f'{cred[:visible]}{mask_char * 4}{cred[-visible:]}'
        return cred

    @staticmethod
    def random_string(string_length=10):
        """Generate a random string of fixed length

        Args:
            string_length (int, optional): The length of the string. Defaults to 10.

        Returns:
            str: A random string
        """
        return ''.join(random.choice(string.ascii_letters) for i in range(string_length))

    def requests_to_curl(self, request, mask_headers=True, mask_patterns=None, verify=True):
        """Return converted PreparedRequest to a curl command.

        Args:
            request (requests.models.PreparedRequest): The response.request object.
            mask_headers (bool, default: True): If True then values for certain header
                key will be masked.
            mask_patterns (list, default: None): A list of patterns if found in headers the value
                will be masked.
            verify (bool, default: True): If False the curl command will include --insecure flag.

        Returns:
            str: The curl command.
        """
        # APP-79 - adding the ability to log request as curl commands
        cmd = ['curl', '-X', request.method]

        # add headers to curl command
        for k, v in sorted(list(dict(request.headers).items())):
            if mask_headers is True:
                patterns = [
                    'authorization',
                    'password',
                    'session',
                    'username',
                    'token',
                ]
                if isinstance(mask_patterns, list):
                    # add user defined mask patterns
                    patterns.extend(mask_patterns)

                for p in patterns:
                    if re.match(rf'.*{p}.*', k, re.IGNORECASE):
                        v = self.printable_cred(v)

                # using gzip in Accept-Encoding with CURL on the CLI produces
                # the warning "Binary output can mess up your terminal."
                if k.lower() == 'accept-encoding':
                    encodings = [e.strip() for e in v.split(',')]
                    for encoding in list(encodings):
                        if encoding in ['gzip']:
                            encodings.remove(encoding)
                    v = ', '.join(encodings)

            cmd.append(f"-H '{k}: {v}'")

        if request.body:
            # add body to the curl command
            body = request.body
            try:
                if isinstance(body, bytes):
                    body = body.decode('utf-8')
                body_data = f'-d "{body}"'
            except Exception:
                temp_file = self.write_temp_binary_file(body)
                body_data = f'--data-binary @{temp_file}'
            cmd.append(body_data)

        if not verify:
            # add insecure flag to curl command
            cmd.append('--insecure')

        # add url to curl command
        cmd.append(request.url)

        return ' '.join(cmd)

    @staticmethod
    def snake_to_camel(snake_string):
        """Convert snake_case to camelCase

        Args:
            snake_string (str): The snake case input string.
        """
        components = snake_string.split('_')
        return components[0] + ''.join(x.title() for x in components[1:])

    @staticmethod
    def standardize_asn(asn: str) -> str:
        """Return the ASN formatted for ThreatConnect."""
        numbers = re.findall('[0-9]+', asn)
        if len(numbers) == 1:
            asn = f'ASN{numbers[0]}'
        return asn

    @staticmethod
    def to_bool(value):
        """Convert value to bool.

        Args:
            value (bool|str): The value to convert to boolean.

        Returns:
            bool: The boolean value
        """
        return str(value).lower() in ['1', 't', 'true', 'y', 'yes']

    def variable_method_name(self, variable):
        """Convert variable name to a valid method name.

        #App:9876:string.operation!String -> string_operation_string

        Args:
            variable (string): The variable name to convert.

        Returns:
            (str): Method name
        """
        method_name = None
        if variable is not None:
            variable = variable.strip()
            if re.match(self.variable_match, variable):
                var = re.search(self.variable_parse, variable)
                variable_name = var.group(3).replace('.', '_').lower()
                variable_type = var.group(4).lower()
                method_name = f'{variable_name}_{variable_type}'
        return method_name

    @property
    def variable_pattern(self):
        """Regex pattern to match and parse a playbook variable."""
        return (
            r'#([A-Za-z]+)'  # match literal (#App) at beginning of String
            r':([\d]+)'  # app id (:7979)
            r':([A-Za-z0-9_\.\-\[\]]+)'  # variable name (:variable_name)
            r'!(StringArray|BinaryArray|KeyValueArray'  # variable type (array)
            r'|TCEntityArray|TCEnhancedEntityArray'  # variable type (array)
            r'|String|Binary|KeyValue|TCEntity|TCEnhancedEntity'  # variable type
            r'|(?:(?!String)(?!Binary)(?!KeyValue)'  # non matching for custom
            r'(?!TCEntity)(?!TCEnhancedEntity)'  # non matching for custom
            r'[A-Za-z0-9_-]+))'  # variable type (custom)
        )

    def write_temp_binary_file(self, content, filename=None):
        """Write content to a temporary file.

        Args:
            content (bytes): The file content.
            filename (str, optional): The filename to use when writing the file.

        Returns:
            str: Fully qualified path name for the file.

        """
        return self.write_temp_file(content, filename, 'wb')

    def write_temp_file(self, content, filename=None, mode='w'):
        """Write content to a temporary file.

        If passing binary data the mode needs to be set to 'wb'.

        Args:
            content (bytes|str): The file content.
            filename (str, optional): The filename to use when writing the file. Defaults to None.
            mode (str, optional): The write mode ('w' or 'wb'). Defaults to w.

        Returns:
            str: Fully qualified path name for the file.
        """
        if filename is None:
            filename = str(uuid.uuid4())
        fqpn = os.path.join(self.temp_path, filename)
        os.makedirs(os.path.dirname(fqpn), exist_ok=True)
        with open(fqpn, mode) as fh:
            fh.write(content)
        return fqpn



# -*- coding: utf-8 -*-
"""ThreatConnect Requests Session"""
import urllib3
from requests import Session, adapters
from urllib3.util.retry import Retry

# disable ssl warning message
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class ExternalSession(Session):
    """ThreatConnect REST API Requests Session for external requests"""

    def __init__(self, tcex):
        """Initialize the Class properties."""
        super().__init__()
        self.tcex = tcex

        # properties
        self._base_url = None
        self._mask_headers = True
        self._mask_patterns = None

        # Update User-Agent
        self.headers.update({'User-Agent': 'TcEx'})

        # Set Proxy
        if self.tcex.default_args.tc_proxy_external:
            self.proxies = self.tcex.proxies
            self.tcex.log.info(
                'Using proxy server for external connectivity '
                f'({self.tcex.default_args.tc_proxy_host}:{self.tcex.default_args.tc_proxy_port}).'
            )

        # Add default Retry
        self.retry()

    @property
    def base_url(self):
        """Return the base url."""
        return self._base_url

    @base_url.setter
    def base_url(self, url):
        """Set base_url."""
        self._base_url = url.strip('/')

    @property
    def mask_headers(self):
        """Return mask patterns."""
        return self._mask_patterns

    @mask_headers.setter
    def mask_headers(self, mask_bool):
        """Return mask patterns."""
        self._mask_headers = mask_bool

    @property
    def mask_patterns(self):
        """Return mask patterns."""
        return self._mask_patterns

    @mask_patterns.setter
    def mask_patterns(self, patterns):
        """Return mask patterns."""
        self._mask_patterns = patterns

    def request(self, method, url, **kwargs):  # pylint: disable=arguments-differ
        """Override request method disabling verify on token renewal if disabled on session."""
        if self.base_url is not None and not url.startswith('https'):
            url = f'{self.base_url}{url}'
        response = super().request(method, url, **kwargs)

        # APP-79 - adding logging of request as curl commands
        self.tcex.log.debug(
            self.tcex.utils.requests_to_curl(
                response.request,
                mask_headers=self.mask_headers,
                mask_patterns=self.mask_patterns,
                verify=self.verify,
            )
        )
        self.tcex.log.debug(f'request url: {response.request.url}')
        self.tcex.log.debug(f'status_code: {response.status_code}')

        return response

    def retry(self, retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 504)):
        """Add retry to Requests Session

        https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.retry.Retry
        """
        retries = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        # mount all https requests
        self.mount('https://', adapters.HTTPAdapter(max_retries=retries))



# -*- coding: utf-8 -*-
"""TcEx Framework Module for working with Datastore in the ThreatConnect Platform."""
# import base64
from datetime import datetime, timedelta
from tcex import TcEx


class Cache:
    """TcEx Cache Class."""

    def __init__(self, tcex: TcEx, domain, data_type, ttl_seconds=None, mapping=None):
        """Initialize class properties.

        Args:
            tcex (object): An instance of TcEx.
            domain (str): A value of “system”, “organization”, or “local”.
            data_type (str): A free form type name for the data.
            ttl_seconds (int, optional): Defaults to None. Number of seconds the cache is valid.
            mapping (dict, optional): Defaults to None. ElasticSearch mappings data.
        """
        self.tcex: TcEx = tcex

        # properties
        self.ttl_seconds = ttl_seconds
        self.ds: TcEx.datastore = self.tcex.datastore(domain, data_type, mapping)

        # Warranty void if any of these are changed.  Don't touch.
        self._cache_data_key = 'cache-data'
        self._cache_date_key = 'cache-date'

    @staticmethod
    def _dt_to_epoch(dt):
        """Convert datetime to epoch seconds."""
        try:
            epoch = dt.timestamp()
        except AttributeError:  # py2
            epoch = (dt - datetime(1970, 1, 1)).total_seconds()
        return epoch

    def add(self, rid, data, raise_on_error=True):
        """Write cache data to the data store.

        Args:
            rid (str): The record identifier.
            data (dict): The record data.
            raise_on_error (bool): If True and not r.ok this method will raise a RunTimeError.
        Returns:
            object : Python request response.
        """
        data = {self._cache_date_key: datetime.utcnow().isoformat(), self._cache_data_key: data}
        return self.ds.post(rid, data, raise_on_error)

    def delete(self, rid, raise_on_error=True):
        """Write cache data to the data store.

        Args:
            rid (str): The record identifier.
            raise_on_error (bool): If True and not r.ok this method will raise a RunTimeError.

        Returns:
            object : Python request response.
        """
        return self.ds.delete(rid, raise_on_error)

    def get(self, rid, data_callback=None, raise_on_error=True):
        """Get cached data from the data store.

        Args:
            rid (str): The record identifier.
            data_callback (callable): A method that will return the data.
            raise_on_error (bool): If True and not r.ok this method will raise a RunTimeError.
        Returns:
            object : Python request response.
        """
        cache_data = None
        ds_data = self.ds.get(rid, raise_on_error=False)
        if ds_data is None:
            ds_data = {'found': False}
        if ds_data is not None:
            expired = False
            if ds_data.get('found') is True:
                cache_data = ds_data.get('_source', {})
                cache_date = cache_data.get(self._cache_date_key)
                if self._is_cache_expired(cache_date):
                    cache_data = None
                    expired = True
                    self.tcex.log.debug(f'Cached data is expired for ({rid}).')

            if expired or ds_data.get('found') is False:
                # when cache is expired or does not exist use callback to get data if possible
                if callable(data_callback):
                    # cache_data = self._encode_data(data_callback(rid))
                    cache_data = data_callback(rid)
                    self.tcex.log.debug(f'Using callback data for ({rid}).')
                    if cache_data:
                        cache_data = self.update(
                            rid, cache_data, raise_on_error
                        )  # update the cache data
            else:
                self.tcex.log.debug(f'Using cached data for ({rid}).')

        return cache_data

    def update(self, rid, data, raise_on_error=True):
        """Write updated cache data to the DataStore.

        Args:
            rid (str): The record identifier.
            data (dict): The record data.
            raise_on_error (bool): If True and not r.ok this method will raise a RunTimeError.

        Returns:
            object : Python request response.
        """
        cache_date = datetime.utcnow().isoformat()
        # cache_data = self._encode_data(data)
        data = {self._cache_date_key: cache_date, self._cache_data_key: data}
        self.ds.put(rid, data, raise_on_error)
        return data

    def _is_cache_expired(self, cached_date):
        cached_datetime = self.tcex.utils.datetime.any_to_datetime(cached_date)
        cache_expires = (cached_datetime + timedelta(seconds=self.ttl_seconds)).timestamp()
        return cache_expires < datetime.utcnow().timestamp()



# -*- coding: utf-8 -*-
"""ThreatConnect Requests Session"""
import base64
import hashlib
import hmac
import time
from urllib.parse import quote

import urllib3
from requests import Session, adapters, auth
from urllib3.util.retry import Retry

from tcex.utils import Utils

# disable ssl warning message
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class HmacAuth(auth.AuthBase):
    """ThreatConnect HMAC Authorization"""

    def __init__(self, access_id, secret_key):
        """Initialize the Class properties."""
        super().__init__()
        self._access_id = access_id
        self._secret_key = secret_key

    def __call__(self, r):
        """Override of parent __call__ method."""
        timestamp = int(time.time())
        signature = f'{r.path_url}:{r.method}:{timestamp}'
        hmac_signature = hmac.new(
            self._secret_key.encode(), signature.encode(), digestmod=hashlib.sha256
        ).digest()
        authorization = f'TC {self._access_id}:{base64.b64encode(hmac_signature).decode()}'
        r.headers['Authorization'] = authorization
        r.headers['Timestamp'] = timestamp
        return r


# APP-102 - adding ability to create sessions with different
# TC credentials for accessing special API endpoints
class TestSession(Session):
    """ThreatConnect REST API Requests Session

    Args:
        api_access_id (str): The ThreatConnect API access id.
        api_secret_key (str): The ThreatConnect API secret key.
        proxy_host (str, kwargs): The proxy hostname.
        proxy_port (int, kwargs): The proxy port number.
        proxy_user (str, kwargs): The proxy username.
        proxy_pass (str, kwargs): The proxy password.
        verify (bool, kwargs): If True SSL verification of the host will be performed.
    """

    def __init__(self, api_access_id, api_secret_key, api_path, logger, **kwargs):
        """Initialize the Class properties."""
        super().__init__()
        self.api_path = api_path
        self.api_access_id = api_access_id
        self.api_secret_key = api_secret_key
        self.log = logger
        self.proxy_host = kwargs.get('proxy_host')
        self.proxy_port = kwargs.get('proxy_port')
        self.proxy_user = kwargs.get('proxy_user')
        self.proxy_pass = kwargs.get('proxy_pass')

        # properties
        self.auth = None
        self.utils = Utils()

        # Update User-Agent
        self.headers.update({'User-Agent': 'TcEx Testing'})

        # Set Proxy
        self.proxies = self.proxy_config

        # Add Retry
        self.retry()

        # Set Verify
        self.verify = kwargs.get('verify', True)

    def _configure_auth(self):
        """Return Auth property for session."""
        if self.api_access_id and self.api_secret_key:
            try:
                # for external Apps or testing Apps locally
                self.auth = HmacAuth(self.api_access_id, self.api_secret_key)
                self.log.debug('Using HMAC authorization.')
            except AttributeError:  # pragma: no cover
                raise RuntimeError('No valid ThreatConnect API credentials provided.')
        else:  # pragma: no cover
            raise RuntimeError('No valid ThreatConnect API credentials provided.')

    @property
    def proxy_config(self):
        """Format the proxy configuration for Python Requests module.

        Generates a dictionary for use with the Python Requests module format
        when proxy is required for remote connections.

        **Example Response**
        ::

            {"http": "http://user:pass@10.10.1.10:3128/"}

        Returns:
           (dictionary): Dictionary of proxy settings
        """
        proxies = {}
        if self.proxy_host is not None and self.proxy_port is not None:

            if self.proxy_username is not None and self.proxy_password is not None:
                proxy_username = quote(self.proxy_username, safe='~')
                proxy_password = quote(self.proxy_password, safe='~')

                # proxy url with auth
                proxy_url = (
                    f'{proxy_username}:{proxy_password}' f'@{self.proxy_host}:{self.proxy_port}'
                )
            else:
                # proxy url without auth
                proxy_url = f'{self.proxy_host}:{self.proxy_port}'
            proxies = {'http': f'http://{proxy_url}', 'https': f'https://{proxy_url}'}
        return proxies

    def request(self, method, url, **kwargs):  # pylint: disable=arguments-differ
        """Override request method disabling verify on token renewal if disabled on session."""
        if self.auth is None:
            self._configure_auth()

        if not url.startswith('https'):
            # automatically add api path to each request
            url = f'{self.api_path}{url}'
        response = super().request(method, url, **kwargs)

        # APP-79 - adding logging of request as curl commands
        self.log.debug(self.utils.requests_to_curl(response.request, verify=self.verify))

        return response

    def retry(self, retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 504)):
        """Add retry to Requests Session

        https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.retry.Retry
        """
        retries = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        # mount all https requests
        self.mount('https://', adapters.HTTPAdapter(max_retries=retries))



# -*- coding: utf-8 -*-
"""App Decorators Module."""
import inspect
import wrapt


class IterateOnArg:
    """Iterate on values stored in ``self.args`` namespace.

    This decorator will iterate over all value of the supplied arg and return results. This feature
    is helpful when processing a single value (String) or array of values (StringArray). If the App
    was provided the arg ``self.args.colors`` with a value of ``['blue', 'green', 'magenta']``, then
    this decorator would call the function 3 times. Each call to the function would pass one value
    from the array. The return values are stored and returned all at once after the last value is
    processed.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        # the args value of "colors" refers to ``self.args.colors``.
        @iterateOnArgs(arg='colors')
        def my_method(colors):
            return colors

        # ** OR **

        # the args value of "colors" refers to ``self.args.colors``.
        @iterateOnArg(arg='colors')
        def my_method(**kwargs):
            return kwargs.get('colors')

    Args:
        arg (str): The arg name from the App which contains the input. This input can be
            a Binary, BinaryArray, KeyValue, KeyValueArray, String, StringArray, TCEntity, or
            TCEntityArray.
        default (str, kwargs): Defaults to None. Default value to pass to method if arg
            value is None. Only supported for String or StringArray.
        fail_enabled (boolean|str, kwargs): Accepts a boolean or string value.  If a boolean value
            is provided that value will control enabling/disabling this feature. A string
            value should reference an item in the args namespace which resolves to a boolean.
            The value of this boolean will control enabling/disabling this feature.
        fail_msg (str, kwargs): The message to log when raising RuntimeError.
        fail_on (list, kwargs): Defaults to None. Fail if data read from Redis is in list.
    """

    def __init__(self, arg, **kwargs):
        """Initialize Class properties"""
        self.arg = arg
        self.default = kwargs.get('default')
        self.fail_enabled = kwargs.get('fail_enabled', False)
        self.fail_msg = kwargs.get('fail_msg', f'Invalid value provided for ({arg}).')
        self.fail_on = kwargs.get('fail_on', [])

    @wrapt.decorator
    def __call__(self, wrapped, instance, args, kwargs):
        """Implement __call__ function for decorator.

        Args:
            wrapped (callable): The wrapped function which in turns
                needs to be called by your wrapper function.
            instance (App): The object to which the wrapped
                function was bound when it was called.
            args (list): The list of positional arguments supplied
                when the decorated function was called.
            kwargs (dict): The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """

        def loop(app, *args, **kwargs):
            """Iterate over data, calling the decorated function for each value.

            Args:
                app (class): The instance of the App class "self".
            """
            # get the signature for the decorated method
            fn_signature = inspect.signature(wrapped, follow_wrapped=True).parameters

            # self.enable (e.g., True or 'fail_on_false') enables/disables this feature
            enabled = self.fail_enabled
            if not isinstance(self.fail_enabled, bool):
                enabled = getattr(app.args, self.fail_enabled)
                if not isinstance(enabled, bool):  # pragma: no cover
                    raise RuntimeError(
                        'The fail_enabled value must be a boolean or resolved to bool.'
                    )
                app.tcex.log.debug(f'Fail enabled is {enabled} ({self.fail_enabled}).')

            # retrieve data from Redis if variable and always return and array.
            results = []
            arg_data = app.tcex.playbook.read(getattr(app.args, self.arg))
            arg_type = app.tcex.playbook.variable_type(getattr(app.args, self.arg))
            if arg_data is None:
                arg_data = [None]
            elif not isinstance(arg_data, list):
                arg_data = [arg_data]

            _index = 0
            _array_length = len(arg_data)
            for ad in arg_data:

                # add "magic" args
                if '_index' in fn_signature:
                    kwargs['_index'] = _index
                if '_array_length' in fn_signature:
                    kwargs['_array_length'] = _array_length

                if ad is None and self.default is not None:
                    # set value passed to method to default if value is None.
                    ad = self.default
                    app.tcex.log.debug(
                        f'replacing null value with provided default value "{self.default}".'
                    )

                # check arg_data against fail_on_values
                if enabled and self.fail_on:
                    if ad in self.fail_on:
                        app.tcex.log.debug(f'Invalid value ({ad}) found for {self.arg}.')
                        app.exit_message = self.fail_msg  # for test case
                        app.tcex.exit(1, self.fail_msg)

                # Add logging for debug/troubleshooting
                if (
                    arg_type not in ['Binary', 'BinaryArray']
                    and app.tcex.log.getEffectiveLevel() <= 10
                ):
                    log_string = str(ad)
                    if len(log_string) > 100:  # pragma: no cover
                        log_string = f'{log_string[:100]} ...'
                    app.tcex.log.debug(f'input value: {log_string}')

                # add results to kwargs
                kwargs[self.arg] = ad
                results.append(wrapped(*args, **kwargs))

                # increment index
                _index += 1
            return results

        return loop(instance, *args, **kwargs)



# -*- coding: utf-8 -*-
"""API Handler Class"""
import os
from logging.handlers import RotatingFileHandler


class RotatingFileHandlerCustom(RotatingFileHandler):
    """Logger handler for ThreatConnect Exchange File logging."""

    def __init__(self, filename, mode='a', maxBytes=0, backupCount=0, encoding=None, delay=0):
        """Customize RotatingFileHandler to create full log path.

        Args:
            filename (str): The name of the logfile.
            mode (str, optional): The write mode for the file. Defaults to 'a'.
            maxBytes (int, optional): The max file size before rotating. Defaults to 0.
            backupCount (int, optional): The maximum # of backup files. Defaults to 0.
            encoding (str, optional): The log file encoding. Defaults to None.
            delay (int, optional): The delay period. Defaults to 0.
        """
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename), exist_ok=True)
        RotatingFileHandler.__init__(self, filename, mode, maxBytes, backupCount, encoding, delay)


# class RotatingFileHandlerFormatter(logging.Formatter):
#     """Custom logging formatter that allows a different format depending on the logging level."""
#
#     def __init__(self):
#         """Initialize formatter parent."""
#         super().__init__(fmt='%(levelno)d: %(msg)s', datefmt=None, style='%')
#
#     def format(self, record):
#         """Format file handle log event according to logging level.
#
#         Args:
#             record (obj): The record to be logged.
#         """
#         # Replace the original format with one customized by logging level
#         self._style._fmt = self.standard_format
#         if record.levelno < 10:  # <= logging.DEBUG
#             self._style._fmt = self.trace_format
#
#         # Call the original formatter class to do the grunt work
#         result = logging.Formatter.format(self, record)
#
#         return result
#
#     @property
#     def standard_format(self):
#         """Return the standard log format"""
#         return (
#             '%(asctime)s - %(name)s - %(levelname)s - %(message)s '
#             '(%(filename)s:%(funcName)s:%(lineno)d:%(threadName)s)'
#         )
#
#     @property
#     def trace_format(self):
#         """Return the standard log format"""
#         return (
#             '%(asctime)s - %(name)s - %(levelname)s - [%(funcName)s:%(lineno)d] %(message)s '
#             '(%(filename)s:%(threadName)s)'
#         )



# -*- coding: utf-8 -*-
"""Bin module for TcEx Framework"""
# flake8: noqa
from .init import Init
from .lib import Lib
from .package import Package
from .test import Test
from .validate import Validate



# -*- coding: utf-8 -*-
"""ThreatConnect Requests Session"""
import base64
import hashlib
import hmac
import time

import urllib3
from requests import Session, adapters, auth
from urllib3.util.retry import Retry

# disable ssl warning message
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class HmacAuth(auth.AuthBase):
    """ThreatConnect HMAC Authorization"""

    def __init__(self, access_id, secret_key):
        """Initialize the Class properties."""
        super().__init__()
        self._access_id = access_id
        self._secret_key = secret_key

    def __call__(self, r):
        """Override of parent __call__ method."""
        timestamp = int(time.time())
        signature = f'{r.path_url}:{r.method}:{timestamp}'
        hmac_signature = hmac.new(
            self._secret_key.encode(), signature.encode(), digestmod=hashlib.sha256
        ).digest()
        authorization = f'TC {self._access_id}:{base64.b64encode(hmac_signature).decode()}'
        r.headers['Authorization'] = authorization
        r.headers['Timestamp'] = timestamp
        return r


class TokenAuth(auth.AuthBase):
    """ThreatConnect Token Authorization"""

    def __init__(self, token):
        """Initialize Class Properties."""
        super().__init__()
        self.token = token

    def __call__(self, r):
        """Override of parent __call__ method."""
        r.headers['Authorization'] = f'TC-Token {self.token.token}'
        return r


class TcSession(Session):
    """ThreatConnect REST API Requests Session"""

    def __init__(self, tcex):
        """Initialize the Class properties."""
        super().__init__()
        self.tcex = tcex

        # properties
        self.args = self.tcex.default_args
        self.auth = None
        self.token = self.tcex.token

        # Update User-Agent
        self.headers.update({'User-Agent': 'TcEx'})

        # Set Proxy
        if self.args.tc_proxy_tc:
            self.proxies = self.tcex.proxies
            self.tcex.log.trace(
                f'Using proxy host {self.args.tc_proxy_host}:'
                f'{self.args.tc_proxy_port} for ThreatConnect API.'
            )

        # Add Retry
        self.retry()

        # Set Verify
        self.verify = self.args.tc_verify

    def _configure_auth(self):
        """Return Auth property for session."""
        # Add ThreatConnect Authorization
        if self._service_app or self._token_available:
            # service Apps only use tokens and playbook/runtime Apps will use token if available
            self.auth = TokenAuth(self.token)
            self.tcex.log.trace('Using token authorization.')
        elif self.args.api_access_id and self.args.api_secret_key:
            try:
                # for external Apps or testing Apps locally
                self.auth = HmacAuth(self.args.api_access_id, self.args.api_secret_key)
                self.tcex.log.trace('Using HMAC authorization.')
            except AttributeError:  # pragma: no cover
                raise RuntimeError('No valid ThreatConnect API credentials provided.')
        else:  # pragma: no cover
            raise RuntimeError('No valid ThreatConnect API credentials provided.')

    @property
    def _service_app(self):
        """Return true if the current App is a service App."""
        return self.tcex.ij.runtime_level.lower() in [
            'apiservice',
            'triggerservice',
            'webhooktriggerservice',
        ]

    @property
    def _token_available(self):
        """Return true if the current App is a service App."""
        return self.token.token is not None and self.token.token_expires is not None

    def request(self, method, url, **kwargs):  # pylint: disable=arguments-differ
        """Override request method disabling verify on token renewal if disabled on session."""
        if self.auth is None:
            self._configure_auth()

        if not url.startswith('https'):
            url = f'{self.args.tc_api_path}{url}'
        response = super().request(method, url, **kwargs)

        # APP-79 - adding logging of request as curl commands
        self.tcex.log.debug(self.tcex.utils.requests_to_curl(response.request, verify=self.verify))
        self.tcex.log.debug(f'request url: {response.request.url}')
        self.tcex.log.debug(f'status_code: {response.status_code}')

        return response

    def retry(self, retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 504)):
        """Add retry to Requests Session

        https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.retry.Retry
        """
        retries = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        # mount all https requests
        self.mount('https://', adapters.HTTPAdapter(max_retries=retries))



# -*- coding: utf-8 -*-
"""TcEx Framework Service module"""
import threading
import time

# from requests import exceptions, get
from requests import Session, exceptions
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


def retry_session(retries=3, backoff_factor=0.8, status_forcelist=(500, 502, 504)):
    """Add retry to Requests Session.

    https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.retry.Retry
    """
    session = Session()
    retries = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    # mount all https requests
    session.mount('https://', HTTPAdapter(max_retries=retries))
    return session


class Tokens:
    """Service methods for customer Service (e.g., Triggers).

    Args:
        token_url (str): The ThreatConnect URL
        verify (bool): A boolean to enable/disable SSL verification
        logger (logging.logger): An pre-configured instance of a logger.
    """

    def __init__(self, token_url, sleep_interval, verify, logger):
        """Initialize the Class properties."""
        self.lock = threading.Lock()
        # tcex logger
        self.log = logger
        # session with retry for token renewal
        self.session = retry_session()
        # shutdown boolean
        self.shutdown = False
        # token monitor sleep interval
        self.sleep_interval = sleep_interval
        # token map for storing keys -> tokens -> threads
        self.token_map = {}
        # base url for token renewal
        self.token_url = token_url
        # amount of seconds to pad before token renewal
        self.token_window = 60
        # ssl verification setting for TC token renewal
        self.verify = verify

        # start token renewal process
        self.token_renewal()

    @property
    def key(self):
        """Return the current key"""
        key = 'MainThread'  # default Python parent thread name
        # self.log.trace(f'in key - thread_name: {self.thread_name}')
        if self.thread_name in self.token_map:
            # for Job, Playbook, and ApiService Apps the key is the thread name.
            key = self.thread_name
        else:
            # for Trigger and Webhook Apps the key is ConfigId.
            # find ConfigId using array of registered thread names.
            for k, d in self.token_map.items():
                if self.thread_name in d.get('thread_names', []):
                    key = k
                    break
            else:  # pragma: no cover
                self.log.trace(f'Thread name not found, defaulting to {key}')
        return key

    @staticmethod
    def printable_token(token):
        """Return a printable token

        Args:
            token (str): The token to print.

        Returns:
            str: The reformatted token.
        """
        if token is not None:
            token = f'{token[:10]}...{token[-10:]}'
        return token

    def register_thread(self, key, thread_name):
        """Register a thread name to a key.

        For Trigger and Webhook Apps multiple threads will share a token registered to a ConfigId.

        Args:
            key (str): The key to use to identify a token.
            thread_name (str): The thread to register to a key.
        """
        self.token_map.setdefault(key, {}).setdefault('thread_names', []).append(thread_name)
        self.log.info(f'Token thread registered -  key: {key}, thread: {thread_name}')

    def register_token(self, key, token, expires):
        """Register a token.

        Args:
            key (str): The key to use to identify a token. Typically a thread name or a config id.
            token (str): The ThreatConnect API token.
            expires (int): The token expiration timestamp.
        """
        if token is None or expires is None:  # pragma: no cover
            self.log.error(
                f'Invalid token data provided - token: {self.printable_token(token)}, '
                f'expires: {expires}.'
            )
            return

        self.token_map[key] = {'thread_names': [], 'token': token, 'token_expires': int(expires)}
        self.log.info(
            f'Token registered - key: {key}, '
            f'token: {self.printable_token(token)}, '
            f'expiration {expires}'
        )

    def renew_token(self, token):
        """Renew expired ThreatConnect Token.

        This method will renew a token and update the token_map with new token and expiration.

        Args:
            token (str): The ThreatConnect API token.
            token_expires (int): The token expiration timestamp.
        """
        api_token_data = {}
        self.log.in_token_renewal = True  # pause API logging

        # log token information
        try:
            params = {'expiredToken': token}
            url = f'{self.token_url}/appAuth'
            r = self.session.get(url, params=params, verify=self.verify)

            if not r.ok:
                err_reason = r.text or r.reason
                err_msg = (
                    f'Token Retry Error. API status code: {r.status_code}, '
                    f'API message: {err_reason}, '
                    f'Token: {self.printable_token(token)}.'
                )
                self.log.error(err_msg)
                raise RuntimeError(1042, err_msg)
        except exceptions.SSLError:  # pragma: no cover
            raise RuntimeError('Token renewal failed with an SSL Error.')

        # process response for token
        try:
            api_token_data = r.json()
        except (AttributeError, ValueError) as e:  # pragma: no cover
            raise RuntimeError(f'Token renewal failed ({e}).')
        finally:
            self.log.in_token_renewal = False

        return api_token_data

    @property
    def thread_name(self):
        """Return the current thread name."""
        return threading.current_thread().name

    @property
    def token(self):
        """Return token for current thread."""
        return self.token_map.get(self.key, {}).get('token')

    @token.setter
    def token(self, token):
        """Set token for current thread."""
        self.token_map.setdefault(self.key, {})['token'] = token

    @property
    def token_expires(self):
        """Return token_expires for current thread."""
        return self.token_map.get(self.key, {}).get('token_expires')

    @token_expires.setter
    def token_expires(self, expires):
        """Set token expires for current thread."""
        self.token_map.setdefault(self.key, {})['token_expires'] = int(expires)

    def token_renewal(self):
        """Start token renewal monitor thread."""
        self.log.debug('Token renewal monitor starting')
        t = threading.Thread(name='token-renewal', target=self.token_renewal_monitor)
        t.daemon = True  # use setter for py2
        t.start()

    def token_renewal_monitor(self):
        """Monitor token expiration and renew when required."""
        while True:
            for key, token_data in dict(self.token_map).items():
                # calculate the time left to sleep
                sleep_seconds = (
                    token_data.get('token_expires') - int(time.time()) - self.token_window
                )
                self.log.debug(
                    f'token status - key: {key}, '
                    f"token: {self.printable_token(token_data.get('token'))}, "
                    f"expires: {token_data.get('token_expires')}, "
                    f'sleep-seconds: {sleep_seconds}'
                )

                if sleep_seconds < 0:
                    # renew token data
                    with self.lock:
                        try:
                            api_token_data = self.renew_token(token_data.get('token'))
                            self.token_map[key]['token'] = api_token_data['apiToken']
                            self.token_map[key]['token_expires'] = int(
                                api_token_data['apiTokenExpires']
                            )
                            self.log.info(
                                f'Token renewed - key: {key}, '
                                f"token: {self.printable_token(api_token_data['apiToken'])}, "
                                f"expires: {api_token_data['apiTokenExpires']}"
                            )
                        except RuntimeError as e:
                            self.log.error(e)
                            try:
                                del self.token_map[key]
                                self.log.error(f'Failed token removed - key: {key}')
                            except KeyError:  # pragma: no cover
                                pass
            time.sleep(self.sleep_interval)
            if self.shutdown:
                break

    def unregister_thread(self, key, thread_name):
        """Unregister a thread name for a key.

        Args:
            key (str): The key to use to identify a token.
            thread_name (str): The thread to unregister from a key.
        """
        try:
            self.token_map[key]['thread_names'].remove(thread_name)
            self.log.info(f'Token thread unregistered -  key: {key}, thread: {thread_name}')
        except (KeyError, ValueError):  # pragma: no cover
            pass

    def unregister_token(self, key):
        """Unregister a token.

        Args:
            key (str): The key used to identify a token.
        """
        try:
            del self.token_map[key]
            self.log.info(f'Token unregistered - key: {key}')
        except KeyError:
            pass



# -*- coding: utf-8 -*-
"""TcEx testing profile Class."""
import json
import math

# import os
import re
import sys
from base64 import b64encode

import colorama as c

# autoreset colorama
c.init(autoreset=True, strip=False)


class Interactive:
    """Testing Profile Interactive Class.

    Args:
        profile (Profile): The profile object to build interactive inputs.
    """

    def __init__(self, profile):
        """Initialize Class properties."""
        self.profile = profile

        # properties
        self._inputs = {
            'optional': {},
            'required': {},
        }
        self._no_selection_text = 'No Selection'
        self._staging_data = {'kvstore': {}}
        # self._user_defaults = None
        self.collect_type_map = {
            'Any': self.collect_string,
            'Binary': self.collect_binary,
            'BinaryArray': self.collect_binary_array,
            'KeyValue': self.collect_key_value,
            'KeyValueArray': self.collect_key_value_array,
            'String': self.collect_string,
            'StringArray': self.collect_string_array,
            'TCEntity': self.collect_tcentity,
            'TCEntityArray': self.collect_tcentity_array,
        }
        self.exit_codes = []
        self.input_type_map = {
            'boolean': self.present_boolean,
            'choice': self.present_choice,
            'keyvaluelist': self.present_key_value_list,
            'multichoice': self.present_multichoice,
            'string': self.present_string,
        }
        # self.user_defaults_filename = os.path.join('tests', '.user_defaults')

    def _default(self, data, data_type=None):  # pylint: disable=unused-argument
        """Return the best option for default."""
        if data.get('type').lower() == 'boolean':
            default = str(data.get('default', 'false')).lower()
        elif data.get('type').lower() == 'choice':
            default = 0
            valid_values = self.profile.ij.expand_valid_values(data.get('validValues', []))
            if data.get('name') == 'tc_action':
                for vv in valid_values:
                    if self.profile.feature.lower() == vv.replace(' ', '_').lower():
                        default = vv
                        break
            else:
                default = data.get('default')
        elif data.get('type').lower() == 'multichoice':
            default = data.get('default')
            if default is not None and isinstance(default, str):
                default = default.split('|')
        else:
            default = data.get('default')
            # if default is None:
            #     # set default from user default file
            #     default = self.user_defaults.get(data.get('name'))
        return default

    def _input_value(self, label, option_text=None):
        """Return user input."""
        # update option text to include help message
        option_text = option_text or ''
        if option_text:
            # add space for cleaness in user display
            option_text = f' {option_text}'

        print(f'{c.Fore.WHITE}[? for help]')
        prompt = f'{c.Fore.MAGENTA}{label}{c.Fore.RESET}{c.Style.BRIGHT}{option_text}: '
        input_value = input(prompt).strip()

        # handle special user inputs
        if input_value == '?':
            self.present_help()
            return self._input_value(label, option_text)

        return input_value

    @staticmethod
    def _split_list(data):
        """Split a list in two "equal" parts."""
        half = math.ceil(len(data) / 2)
        return data[:half], data[half:]

    def add_input(self, name, data, value):
        """Add an input to inputs."""
        if data.get('required', False):
            self._inputs['required'].setdefault(name, value)
        else:
            self._inputs['optional'].setdefault(name, value)

    # def add_user_default(self, key, value, data_type=None):
    #     """Add data to user default."""
    #     self.user_defaults.setdefault(self.profile.feature, {})
    #     if data_type is None:
    #         self.user_defaults[self.profile.feature][key] = value
    #     else:
    #         # store the value under the appropriate data type
    #         self.user_defaults[self.profile.feature].setdefault(key, {})
    #         self.user_defaults[self.profile.feature][key].setdefault(data_type, value)

    #     if self.user_defaults.get('base') is None:
    #         self.user_defaults['base'] = self.user_defaults[self.profile.feature]

    def add_staging_data(self, name, type_, value):
        """Create staging data and return variable value.

        Args:
            name (str): The name of the input.
            type_ (str): The type of input (Binary, StringArray, etc.)
            value (str): The value to write in the staging data.

        Returns:
            [type]: [description]
        """
        arg_value = value
        if (
            self.profile.ij.runtime_level.lower() not in ['triggerservice', 'webhooktriggerservice']
            and value is not None
        ):
            arg_value = self.profile.ij.create_variable(name, type_)
            self._staging_data['kvstore'].setdefault(arg_value, value)

        return arg_value

    def collect_binary(self, **kwargs):
        """Collect binary data

        Args:
            default (str, kwargs): The default value if no value provided by user.
            feedback (book, kwargs): If True user feedback will be printed.
            option_text (str, kwargs): The text shown to the user.
            required (str, kwargs): If True the user cannot continue until they provide a value.
        """
        input_value = self._input_value('Input', kwargs.get('option_text'))
        if not input_value:
            # if no default value and required force user to input again
            if kwargs.get('default') is None and kwargs.get('required') is True:
                self.print_required()
                return self.collect_binary(**kwargs)

        if input_value not in [None, '']:
            input_data = b64encode(input_value.encode()).decode()
            feedback = f'{input_value} -> ({input_data})'
        else:
            input_data = kwargs.get('default')
            feedback = input_data

        # print user feedback
        if kwargs.get('feedback', True):
            self.print_feedback(feedback)

        return input_data

    def collect_binary_array(self, **kwargs):
        """Collect binary array data

        Args:
            required (str, kwargs): If True the user cannot continue until they provide a value.
        """
        input_values = []
        required = kwargs.get('required', False)
        while True:
            input_value = self.collect_binary(feedback=False, required=required)
            if not input_value:
                break
            input_values.append(input_value)
            required = False  # only the first input is required

        if not input_values:
            # return None to ensure data doesn't get added to inputs
            input_values = None

        # print user feedback
        self.print_feedback(input_values)

        return input_values

    def collect_boolean(self, **kwargs):
        """Collect binary data

        Args:
            default (str, kwargs): The default value if no value provided by user.
            option_text (str, kwargs): The text shown to the user.
        """
        input_value = self._input_value('Input', kwargs.get('option_text'))
        if input_value == '':
            input_value = kwargs.get('default')

        if str(input_value).lower() not in ['0', 'f', 'false', '1', 't', 'true']:
            self.print_invalid_bool()
            return self.collect_boolean(**kwargs)

        # covert input value to a proper boolean
        input_value = self.profile.utils.to_bool(input_value)

        # print user feedback
        self.print_feedback(input_value)

        return input_value

    def collect_choice(self, **kwargs):
        """Collect choice data

        Args:
            default (str, kwargs): The default value if no value provided by user.
            option_text (str, kwargs): The text shown to the user.
            required (str, kwargs): If True the user cannot continue until they provide a value.
            valid_values (str, kwargs): A list of valid values
        """
        # collect input value from user and set default if required
        input_value = self._input_value('Choice', kwargs.get('option_text')) or kwargs.get(
            'default'
        )

        # ensure input value is provided when input is required
        if input_value is None and kwargs.get('required') is True:
            self.print_required()
            return self.collect_choice(**kwargs)

        # if input value is None then there is not need to continue
        if input_value is None:
            return input_value

        # set valid values
        valid_values = kwargs.get('valid_values', [])

        # convert to int or recollect input
        try:
            input_value = int(input_value)
        except ValueError:
            self.print_invalid_index(f'0-{len(valid_values)}')
            return self.collect_choice(**kwargs)

        # ensure input value is valid
        valid_index_values = [i for i, _ in enumerate(valid_values)]
        # valid_index_values = list(range(0, len(valid_values) - 1))
        if input_value not in valid_index_values:
            self.print_invalid_index(f'0-{len(valid_values)}')
            return self.collect_choice(**kwargs)

        # using index value provided by user, set value to valid value
        input_value = valid_values[input_value]
        if input_value == self._no_selection_text:
            # special case for when user select no selection
            input_value = None

        # print user feedback
        if kwargs.get('feedback', True):
            self.print_feedback(input_value)

        return input_value

    def collect_exit_code(self, **kwargs):
        """Collect exit codes."""
        input_value = self._input_value('Code', kwargs.get('option_text'))

        if input_value != '':
            try:
                input_value = int(input_value)
            except ValueError:
                self.print_invalid_exit_code()
                return self.collect_exit_code(**kwargs)

            if input_value not in [0, 1, 3]:
                self.print_invalid_exit_code()
                return self.collect_exit_code(**kwargs)

        return input_value

    def collect_exit_codes(self, **kwargs):
        """Collect exit codes."""
        input_values = []
        while True:
            input_value = self.collect_exit_code(**kwargs)
            if input_value == '':
                break
            input_values.append(input_value)

        if not input_values:
            # return None to ensure data doesn't get added to inputs
            input_values = [0]

        # print user feedback
        self.print_feedback(input_values)

        return input_values

    def collect_key_value(self, **kwargs):
        """Collect key value data"""
        input_value = None
        key = self._input_value('Key', option_text=kwargs.get('option_text'))

        # ensure input value is provided when input is required
        if key == '' and kwargs.get('required') is True:
            self.print_required()
            return self.collect_key_value(**kwargs)

        if key != '':
            value = self._input_value('Value')
            input_value = {'key': key, 'value': value}
        else:
            input_value = kwargs.get('default')

        # print user feedback
        if kwargs.get('feedback', True):
            self.print_feedback(input_value)

        return input_value

    def collect_key_value_array(self, **kwargs):
        """Collect key value array data"""
        input_values = []
        required = kwargs.get('required')
        while True:
            input_value = self.collect_key_value(
                default=kwargs.get('default'),
                feedback=False,
                option_test=kwargs.get('option_text'),
                required=required,
            )
            if not input_value:
                break
            input_values.append(input_value)
            required = False

        if not input_values:
            # return None to ensure data doesn't get added to inputs
            input_values = None

        # print user feedback
        self.print_feedback(input_values)

        return input_values

    def collect_multichoice(self, **kwargs):
        """Collect multichoice data"""
        input_values = []
        required = kwargs.get('required', False)
        while True:
            input_value = self.collect_choice(
                feedback=False,
                # option_text=kwargs.get('option_text'),
                required=required,
                valid_values=kwargs.get('valid_values'),
            )
            if not input_value:
                break
            input_values.append(input_value)
            required = False

        input_values = list(set(input_values))
        if input_values:
            # format multichoice value as pipe delimited string
            input_values = '|'.join(input_values)
        else:
            # return None to ensure data doesn't get added to inputs
            input_values = None

        # print user feedback
        self.print_feedback(input_values)

        return input_values

    def collect_string(self, **kwargs):
        """Collect string data

        Args:
            option_text (str, kwargs): The text shown to the user.
            default (str, kwargs): The default value if no value provided by user.
        """
        input_value = self._input_value('Input', kwargs.get('option_text', ''))
        if not input_value:
            input_value = kwargs.get('default')

        if input_value is None and kwargs.get('required', False) is True:
            self.print_required()
            return self.collect_string(**kwargs)

        # print user feedback
        if kwargs.get('feedback', True):
            self.print_feedback(input_value)

        # APP-622 - handle null/None values
        if input_value == 'null':
            input_value = None
        elif input_value in ['"null"', "'null'"]:
            input_value = 'null'

        return input_value

    def collect_string_array(self, **kwargs):
        """Collect string array data"""
        input_values = []
        required = kwargs.get('required', False)
        while True:
            input_value = self.collect_string(feedback=False, required=required)
            if not input_value:
                break
            input_values.append(input_value)
            required = False

        if not input_values:
            # return None to ensure data doesn't get added to inputs
            input_values = None

        # print user feedback
        self.print_feedback(input_values)

        return input_values

    def collect_tcentity(self, **kwargs):
        """Collect tcentity data"""
        input_value = None
        id_ = self._input_value('ID')
        if id_:
            value = self._input_value('Value')
            type_ = self._input_value('Type')
            input_value = {'id': id_, 'value': value, 'type': type_}

        if input_value is None and kwargs.get('required', False) is True:
            self.print_required()
            return self.collect_tcentity(**kwargs)

        # print user feedback
        if kwargs.get('feedback', True):
            self.print_feedback(input_value)

        return input_value

    def collect_tcentity_array(self, **kwargs):
        """Collect tcentity array data"""
        input_values = []
        required = kwargs.get('required', False)
        while True:
            input_value = self.collect_tcentity(feedback=False, required=required)
            if not input_value:
                break
            input_values.append(input_value)
            required = False

        if not input_values:
            # return None to ensure data doesn't get added to inputs
            input_values = None

        # print user feedback
        self.print_feedback(input_values)

        return input_values

    @property
    def inputs(self):
        """Return inputs dict."""
        return self._inputs

    def present(self):
        """Present interactive menu to build profile."""

        def params_data():
            # handle non-layout and layout based App appropriately
            if self.profile.lj.has_layout:
                # using inputs from layout.json since they are required to be in order
                # (display field can only use inputs previously defined)
                for name in self.profile.lj.params_dict:
                    # get data from install.json based on name
                    data = self.profile.ij.params_dict.get(name)
                    yield name, data

                # hidden fields will not be in layout.json so they need to be include manually
                for name, data in self.profile.ij.filter_params_dict(hidden=True).items():
                    yield name, data
            else:
                for name, data in self.profile.ij.params_dict.items():
                    yield name, data

        inputs = {}
        for name, data in params_data():
            if data.get('serviceConfig'):
                # inputs that are serviceConfig are not applicable for profiles
                continue

            if not data.get('hidden'):
                # each input will be checked for permutations if the App has layout and not hidden
                if not self.profile.permutations.validate_input_variable(name, inputs):
                    continue

            # present the input
            value = self.input_type_map.get(data.get('type').lower())(name, data)

            # update inputs
            inputs[name] = value

        self.present_exit_code()

    def present_boolean(self, name, data):
        """Build a question for boolean input."""
        # print header information
        self.print_header(data)

        default = self._default(data)
        valid_values = ['true', 'false']

        option_default = 'false'
        option_text = ''
        options = []
        for v in valid_values:
            if v.lower() == default.lower():
                option_default = v
                v = f'[{v}]'
            options.append(v)
        option_text = f'''({'/'.join(options)})'''

        value = self.collect_boolean(default=option_default, option_text=option_text)

        # add input
        self.add_input(name, data, value)

        return value

    def present_choice(self, name, data):
        """Build a question for choice input."""
        # print header information
        self.print_header(data)

        default = self._default(data)
        option_index = 0
        valid_values = self.profile.ij.expand_valid_values(data.get('validValues', []))
        if data.get('required', False) is False:
            # add option to invalidate defaults
            valid_values.insert(0, self._no_selection_text)

        # default value needs to be converted to index
        if default:
            try:
                option_index = valid_values.index(default)
            except ValueError:
                # if "magic" variable (e.g., ${GROUP_TYPES}) was not expanded then use index 0.
                # there is no way to tell if the default value is be part of the expansion.
                if any([re.match(r'^\${.*}$', v) for v in valid_values]):
                    option_index = 0
                else:
                    print(
                        f'''{c.Fore.RED}Invalid value of ({default}) for {data.get('name')}, '''
                        'check that default value and validValues match in install.json.'
                    )
                    sys.exit()
        option_text = f'[{option_index}]'

        # build options list to display to the user in two columns
        options = []
        for i, v in enumerate(valid_values):
            options.append(f'{i}. {v}')

        # display options list into two columns
        left, right = self._split_list(options)
        for i, _ in enumerate(left):
            ld = left[i]
            try:
                rd = right[i]
            except IndexError:
                rd = ''
            print(f'{ld:40} {rd:40}')

        # collect user input
        value = self.collect_choice(
            default=option_index, option_text=option_text, valid_values=valid_values
        )

        # add input
        self.add_input(name, data, value)

        return value

    def present_data_types(self, data_types, required=False):
        """Present data types options.

        Args:
            data_types (list): A list of optional data types.
            required (bool): If False the no selection option will be added.
        """
        if 'Any' in data_types:
            data_types = [
                'Binary',
                'BinaryArray',
                'KeyValue',
                'KeyValueArray',
                'String',
                'StringArray',
                'TCEntity',
                'TCEntityArray',
            ]

        # add option to not select an index value if input is not required
        if required is False:
            data_types.insert(0, self._no_selection_text)

        # build options list to display to the user in two columns
        options = []
        for i, v in enumerate(data_types):
            options.append(f'{i}. {v}')

        left, right = self._split_list(options)
        for i, _ in enumerate(left):
            ld = left[i]
            try:
                rd = right[i]
            except IndexError:
                rd = ''
            print(f'{ld:40} {rd:40}')

        index = self._input_value('Type', '[0]') or 0

        try:
            data_type = data_types[int(index)]
        except (IndexError, TypeError, ValueError):
            print(
                f'{c.Fore.RED}Invalid index of {index} provided. '
                f'Please provide a integer between 0-{len(data_types) - 1}'
            )
            sys.exit(1)

        return data_type

    def present_exit_code(self):
        """Provide user input for exit code."""
        self.print_header({'label': 'Exit Codes'})
        self.exit_codes = list(set(self.collect_exit_codes(default=[0], option_text='[0]')))

    @staticmethod
    def present_help():
        """Provide user help information."""
        print(
            f'{c.Fore.CYAN}For String type inputs: \n'
            ' * A value of null will be treated as an actual null value.\n'
            ' * Using "null" or \'null\' to insert a string of null.\n'
        )
        print(f'{c.Fore.CYAN}When done entering array data press enter to continue.')

    def present_key_value_list(self, name, data):
        """Build a question for key value list input."""
        # print header information
        self.print_header(data)

        # the default value from install.json or user_data
        default = self._default(data)  # array of default values

        # collect input
        input_data = self.collect_key_value_array(default=default, required=data.get('required'))

        # create variable
        variable = self.add_staging_data(name, 'KeyValueArray', input_data)

        # add input to args
        self.add_input(name, data, variable)

        # user feedback
        feedback_data = input_data
        if input_data is not None:
            feedback_data = json.dumps(feedback_data)

        # # update default
        # if default is None:
        #     self.add_user_default(name, input_data)

        return variable

    def present_multichoice(self, name, data):
        """Build a question for choice input."""
        # print header information
        self.print_header(data)

        default = self._default(data)  # array of default values
        option_indexes = [0]
        valid_values = self.profile.ij.expand_valid_values(data.get('validValues', []))
        if data.get('required', False) is False:
            # add option to invalidate defaults
            valid_values.insert(0, self._no_selection_text)

        # default values will be return as an array (e.g., one|two -> ['one'. 'two']).
        # using the valid values array we can look up these values to show as default in input.
        if default:
            option_indexes = []
            for d in default:
                try:
                    option_indexes.append(valid_values.index(d))
                except ValueError:
                    # if "magic" variable (e.g., ${GROUP_TYPES}) was not expanded then skip value.
                    # there is no way to tell if the default value is be part of the expansion.
                    if any([re.match(r'^\${.*}$', v) for v in valid_values]):
                        continue

                    print(
                        f'''{c.Fore.RED}Invalid value of ({d}) for {data.get('name')}, check '''
                        'that default value(s) and validValues match in install.json.'
                    )
                    sys.exit()
        option_text = f''' [{','.join([str(v) for v in option_indexes])}]'''

        # build options list to display to the user in two columns
        options = []
        for i, v in enumerate(valid_values):
            options.append(f'{i}. {v}')

        # display options list into two columns
        left, right = self._split_list(options)
        for i, _ in enumerate(left):
            ld = left[i]
            try:
                rd = right[i]
            except IndexError:
                rd = ''
            print(f'{ld:40} {rd:40}')

        # collect user input
        values = self.collect_multichoice(
            default=option_indexes,
            option_text=option_text,
            required=data.get('required'),
            valid_values=valid_values,
        )

        # add input
        self.add_input(name, data, values)

        return values

    def present_string(self, name, data):
        """Build a question for boolean input."""
        # display header information
        self.print_header(data)

        # use playbook data types to determine what input to provide (default to String)
        data_type = data.get('playbookDataType', ['String'])[0]
        if len(data.get('playbookDataType', [])) > 1 or data_type.lower() == 'any':
            data_type = self.present_data_types(
                data.get('playbookDataType'), required=data.get('required', False)
            )

        # no need to proceed if there is not valid data type selected.
        if data_type == self._no_selection_text:
            self.add_input(name, data, None)
            self.print_feedback('null')
            return None

        # the default value from install.json or user_data
        default = self._default(data, data_type)

        option_text = ''
        if default is not None:
            option_text = f'[{default}]'

        # use data_type to properly format collection input
        input_value = self.collect_type_map[data_type](
            default=default, option_text=option_text, required=data.get('required', False)
        )

        # add staging data and get variable name
        variable = self.add_staging_data(name, data_type, input_value)

        # add input
        self.add_input(name, data, variable)

        # # update default
        # if default is None:
        #     if len(data.get('playbookDataType', [])) > 1 or data_type.lower() == 'any':
        #         # for inputs that take multiple types we need to store user default with the type
        #         self.add_user_default(name, input_value, data_type)
        #     else:
        #         self.add_user_default(name, input_value)

        return variable

    @staticmethod
    def print_feedback(feedback_value):
        """Print the value used."""
        print(f'Using value: {c.Fore.GREEN}{feedback_value}\n')

    @staticmethod
    def print_header(data):
        """Enrich the header with metadata."""

        def _print_metadata(title, value):
            """Print the title and value"""
            print(f'{c.Fore.CYAN}{title!s:<22}: {c.Fore.RESET}{c.Style.BRIGHT}{value}')

        label = data.get('label', 'NO LABEL')
        print(f'\n{c.Fore.GREEN}{label}')

        # type
        _print_metadata('Type', data.get('type'))

        # default
        default = data.get('default')
        if default:
            _print_metadata('Default', default)

        # note
        note = data.get('note', '')[:200]
        if note:
            _print_metadata('Note', note)

        # required
        _print_metadata('Required', str(data.get('required', False)).lower())

        # hidden
        if data.get('hidden'):
            _print_metadata('Hidden', 'true')

        # Input Types
        pbt = ','.join(data.get('playbookDataType', []))
        if pbt:
            _print_metadata('Playbook Data Types', pbt)

        vv = ','.join(data.get('validValues', []))
        if vv:
            _print_metadata('Valid Values', vv)

        print('-' * 50)

    @staticmethod
    def print_invalid_bool():
        """Print a invalid bool error."""
        print(f'{c.Fore.RED}The provided value is not a boolean value (true/false).\n')

    @staticmethod
    def print_invalid_exit_code():
        """Print a invalid exit code error."""
        print(f'{c.Fore.RED}The provided value is not a valid exit code (0, 1).\n')

    @staticmethod
    def print_invalid_index(range_):
        """Print a invalid index error."""
        print(
            f'{c.Fore.RED}The provided index value is not '
            f'valid, please select a valid value between {range_}.\n'
        )

    @staticmethod
    def print_required():
        """Print a required error."""
        print(f'{c.Fore.RED}This input is required, please enter an appropriate value.\n')

    @property
    def staging_data(self):
        """Return staging data dict."""
        return self._staging_data

    # @property
    # def user_defaults(self):
    #     """Return user defaults"""
    #     if self._user_defaults is None:
    #         user_defaults = {}
    #         if os.path.isfile(self.user_defaults_filename):
    #             with open(self.user_defaults_filename, 'r') as fh:
    #                 user_defaults = json.load(fh)

    #         # use feature defaults
    #         self._user_defaults = user_defaults.get(self.profile.feature)
    #         if self._user_defaults is None:
    #             # use base defaults if not feature defaults found
    #             self._user_defaults = user_defaults.get('base', {})

    #     return self._user_defaults



# -*- coding: utf-8 -*-
"""ReadArg App Decorators."""
import wrapt


class ReadArg:
    """Read value of App arg resolving any playbook variables.

    This decorator will read the args from Redis, if it is a playbook variable, and pass the
    resolved value to the function.

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        @ReadArg(arg='color', default='red')
        @ReadArg(arg='number')
        @ReadArg(
            arg='fruits',
            array=True,
            fail_on=[None, []],
            fail_enabled=True,
            fail_msg='Invalid input for fruits'
        )
        def my_method(self, color, number, fruits):
            print('color', color)
            print('number', number)
            print('fruit', fruit)

        # ** OR **

        @ReadArg(arg='color', default='red')
        @ReadArg(arg='number')
        @ReadArg(
            arg='fruits',
            array=True,
            fail_on=[None, []],
            fail_enabled=True,
            fail_msg='Invalid input for fruits'
        )
        def my_method(self, **kwargs):
            print('color', kwargs.get('color'))
            print('number', kwargs.get('number')
            print('fruit', kwargs.get('fruit')

    Args:
        arg (str): The arg name from the App which contains the input. This input can be
            a Binary, BinaryArray, KeyValue, KeyValueArray, String, StringArray, TCEntity, or
            TCEntityArray.
        array (bool, kwargs): Defaults to False. If True the arg value will always be
            returned as an array.
        default (str, kwargs): Defaults to None. Default value to pass to method if arg
            value is None. Only supported for String or StringArray.
        fail_enabled (boolean|str, kwargs): Accepts a boolean or string value.  If a boolean value
            is provided that value will control enabling/disabling this feature. A string
            value should reference an item in the args namespace which resolves to a boolean.
            The value of this boolean will control enabling/disabling this feature.
        fail_msg (str, kwargs): The message to log when raising RuntimeError.
        fail_on (list, kwargs): Defaults to None. Fail if data read from Redis is in list.
        group_values (bool, kwargs): Defaults to False. If True, return a list of group names from
            the given argument.
        group_ids (bool, kwargs): Defaults to False. If True, return a list of group ids from the
            given argument.
        indicator_values (bool, kwargs): Defaults to False. If True, return a list of
            indicator values from the given argument (e.g. ["foo.example.com", "bar.example.com"]).
        strip_values (bool, kwargs): Defaults to False.  If True, automatically strips whitespace
            from arg value.  Not valid if group_values, group_ids, or indicator_values is True.
    """

    def __init__(self, arg, **kwargs):
        """Initialize Class properties"""
        self.arg = arg
        self.array = kwargs.get('array')
        self.default = kwargs.get('default')
        self.fail_enabled = kwargs.get('fail_enabled', True)
        self.fail_msg = kwargs.get('fail_msg', f'Invalid value provided for ({arg}).')
        self.fail_on = kwargs.get('fail_on', [])
        self.embedded = kwargs.get('embedded', True)
        self.indicator_values = kwargs.get('indicator_values', False)
        self.group_values = kwargs.get('group_values', False)
        self.group_ids = kwargs.get('group_ids', False)
        self.strip_values = kwargs.get('strip_values', False)

    @wrapt.decorator
    def __call__(self, wrapped, instance, args, kwargs):
        """Implement __call__ function for decorator.

        Args:
            wrapped (callable): The wrapped function which in turns
                needs to be called by your wrapper function.
            instance (App): The object to which the wrapped
                function was bound when it was called.
            args (list): The list of positional arguments supplied
                when the decorated function was called.
            kwargs (dict): The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """

        def read(app, *args, **kwargs):
            """Retrieve/Read data from Redis.

            Args:
                app (class): The instance of the App class "self".
            """
            # self.enable (e.g., True or 'fail_on_false') enables/disables this feature
            enabled = self.fail_enabled
            if not isinstance(self.fail_enabled, bool):
                enabled = getattr(app.args, self.fail_enabled)
                if not isinstance(enabled, bool):  # pragma: no cover
                    raise RuntimeError(
                        'The fail_enabled value must be a boolean or resolved to bool.'
                    )
                app.tcex.log.debug(f'Fail enabled is {enabled} ({self.fail_enabled}).')

            try:
                # read arg from namespace
                arg = getattr(app.args, self.arg)
            except AttributeError:
                if enabled:
                    app.tcex.log.error(f'Arg {self.arg} was not found in Arg namespace.')
                    app.exit_message = self.fail_msg  # for test cases
                    app.tcex.exit(1, self.fail_msg)
                else:
                    # add results to kwargs
                    kwargs[self.arg] = self.default
                    return wrapped(*args, **kwargs)

            # retrieve data from Redis and call decorated function
            if self.indicator_values:
                arg_data = app.tcex.playbook.read_indicator_values(arg)
            elif self.group_values:
                arg_data = app.tcex.playbook.read_group_values(arg)
            elif self.group_ids:
                arg_data = app.tcex.playbook.read_group_ids(arg)
            else:
                arg_data = app.tcex.playbook.read(arg, self.array, embedded=self.embedded)
                if self.strip_values and arg_data is not None:
                    variable_type = app.tcex.playbook.variable_type(arg)
                    if variable_type == 'String':
                        arg_data = arg_data.strip()
                    elif variable_type == 'StringArray':
                        arg_data = [s.strip() for s in arg_data]
                    else:
                        app.tcex.log.warn(
                            f'strip_values is enabled but variable is of type {variable_type}, '
                            f'only String and StringArray variables can be stripped.'
                        )

            if self.default is not None and arg_data is None:
                arg_data = self.default
                app.tcex.log.debug(
                    f'replacing null value with provided default value "{self.default}".'
                )

            # check arg_data against fail_on_values
            if enabled and self.fail_on:
                if arg_data in self.fail_on:
                    app.tcex.log.error(f'Invalid value ({arg_data}) found for {self.arg}.')
                    app.exit_message = self.fail_msg  # for test cases
                    app.tcex.exit(1, self.fail_msg)

            # add results to kwargs
            kwargs[self.arg] = arg_data

            # Add logging for debug/troubleshooting
            arg_type = app.tcex.playbook.variable_type(getattr(app.args, self.arg))
            if arg_type not in ['Binary', 'BinaryArray'] and app.tcex.log.getEffectiveLevel() <= 10:
                # only log variable data to prevent logging keychain data
                if app.tcex.playbook.is_variable(arg):
                    log_string = str(arg_data)
                    if len(log_string) > 100:  # pragma: no cover
                        log_string = f'{log_string[:100]} ...'
                    app.tcex.log.debug(f'input value: {log_string}')

            return wrapped(*args, **kwargs)

        return read(instance, *args, **kwargs)



# -*- coding: utf-8 -*-
"""TcEx Framework Service module"""
import base64
import json
import ssl
import threading
import time
import traceback
import uuid
from datetime import datetime
from io import StringIO

import paho.mqtt.client as mqtt


class Services:
    """Service methods for customer Service (e.g., Triggers).

    Args:
        tcex (object): Instance of TcEx.
    """

    def __init__(self, tcex):
        """Initialize the Class properties."""
        self.tcex = tcex
        self.log = self.tcex.log

        # properties
        self._client = None
        self._connected = False
        self._mqtt_client = None
        self._ready = False
        self._redis_client = None
        self._start_time = datetime.now()
        self.configs = {}
        self.config_thread = None
        self.heartbeat_max_misses = 3
        self.heartbeat_miss_count = 0
        self.heartbeat_sleep_time = 1
        self.heartbeat_watchdog = 0
        self.ready = False
        self.shutdown = False

        # default metrics per service type
        if self.tcex.ij.runtime_level.lower() in ['triggerservice', 'webhooktriggerservice']:
            self._metrics = {'Active Playbooks': 0, 'Errors': 0, 'Hits': 0, 'Misses': 0}
        else:
            self._metrics = {'Errors': 0, 'Requests': 0, 'Responses': 0}

        # config callbacks
        self.api_event_callback = None
        self.create_config_callback = None
        self.delete_config_callback = None
        self.shutdown_callback = None
        self.webhook_event_callback = None

    def _tcex_testing(self, session_id, trigger_id):
        """Write data required for testing framework to Redis."""
        if self.tcex.args.tcex_testing_context is not None:
            _context_tracker = (
                self.redis_client.hget(self.tcex.args.tcex_testing_context, '_context_tracker')
                or '[]'
            )
            _context_tracker = json.loads(_context_tracker)
            _context_tracker.append(session_id)
            self.redis_client.hset(
                self.tcex.args.tcex_testing_context,
                '_context_tracker',
                json.dumps(_context_tracker),
            )
            self.redis_client.hset(session_id, '_trigger_id', trigger_id)

            # log
            self.tcex.log.info(f'Wrote context {session_id} to _context_tracker')
            self.tcex.log.info(f'Wrote trigger id {trigger_id} to _trigger_id')

    def add_metric(self, label, value):
        """Add a metric.

        Metrics are reported in heartbeat message.

        Args:
            label (str): The metric label (e.g., hits) to add.
            value (int|str): The value for the metric.
        """
        self._metrics[label] = value

    def create_config(self, trigger_id, config, message, status):
        """Add config item to service config object.

        Args:
            trigger_id (int): The trigger ID for the current config.
            config (dict): The config for the current trigger.
            message (str): A simple message for the action.
            status (str): The passed/fail status for the App handling of config.
        """
        try:
            # add config to configs
            self.configs[trigger_id] = config

            # send ack response
            response = {
                'command': 'Acknowledged',
                'message': message,
                'status': status,
                'type': 'CreateConfig',
                'triggerId': trigger_id,
            }
            self.publish(json.dumps(response))
        except Exception as e:
            self.tcex.log.error(f'Could not create config for Id {trigger_id} ({e}).')
            self.tcex.log.trace(traceback.format_exc())

    def delete_config(self, trigger_id, message, status):
        """Delete config item from config object.

        Args:
            trigger_id (int): The trigger ID for the current config.
            message (str): A simple message for the action.
            status (str): The passed/fail status for the App handling of config.
        """
        try:
            # delete config from configs dict
            del self.configs[trigger_id]

            # send ack response
            response = {
                'command': 'Acknowledged',
                'message': message,
                'status': status,
                'type': 'DeleteConfig',
                'triggerId': trigger_id,
            }
            self.publish(json.dumps(response))
        except Exception as e:
            self.tcex.log.error(f'Could not delete config for Id {trigger_id} ({e}).')

    def fire_event(self, callback, **kwargs):
        """Trigger a FireEvent command.

        Args:
            callback (callable): The trigger method in the App to call.
            trigger_ids (list, kwargs): A list of trigger ids to trigger.
        """
        if not callable(callback):
            raise RuntimeError('Callback method (callback) is not a callable.')

        # get developer passed trigger_ids
        trigger_ids = kwargs.pop('trigger_ids', None)

        for trigger_id, config in list(self.configs.items()):
            if trigger_ids is not None and trigger_id not in trigger_ids:
                # skip config that don't match developer provided trigger ids
                continue

            try:
                self.tcex.log.trace(f'triggering callback for config id: {trigger_id}')
                # get a session_id specifically for this thread
                session_id = self.session_id(trigger_id)

                # only required for testing in tcex framework
                self._tcex_testing(session_id, trigger_id)

                # get instance of playbook specifically for this thread
                outputs = config.get('tc_playbook_out_variables') or []
                if isinstance(outputs, str):
                    outputs = outputs.split(',')

                playbook = self.tcex.pb(context=session_id, output_variables=outputs)

                self.tcex.log.info(f'Trigger Session ID: {session_id}')

                args = (callback, playbook, trigger_id, config)
                # current thread has session_id as name
                self.message_thread(session_id, self.fire_event_trigger, args, kwargs)
            except Exception:
                self.tcex.log.trace(traceback.format_exc())

    def fire_event_trigger(self, callback, playbook, trigger_id, config, **kwargs):
        """Fire event for trigger.

        Args:
            callback (callable): The App callback method for firing an event.
            playbook (tcex.Playbook): A configure playbook instance for using to interact with
                KvStore.
            trigger_id (int): The current trigger Id.
            config (dict): A dict containing the configuration information.
        """
        # tokens are registered with a key, in this case the key is the trigger_id. each thread
        # is required to be registered to a key to use during lookup.
        self.tcex.token.register_thread(trigger_id, self.thread_name)

        # add a logging handler for this thread
        self.tcex.logger.add_thread_file_handler(
            name=self.thread_name,
            filename=self.session_logfile,
            level=self.tcex.default_args.tc_log_level,
            path=self.tcex.default_args.tc_log_path,
        )
        self.tcex.log.info(f'Handling fire event trigger ({self.thread_name})')

        try:
            if callback(playbook, trigger_id, config, **kwargs):
                # self.tcex.log.info(f'Trigger ID {trigger_id} hit.')
                self.increment_metric('Hits')
                self.fire_event_publish(trigger_id, self.thread_name)
            else:
                self.increment_metric('Misses')
                self.tcex.log.info(f'Trigger ID {trigger_id} missed.')
        except Exception as e:
            self.increment_metric('Errors')
            self.tcex.log.error(f'The callback method encountered and error ({e}).')
            self.tcex.log.trace(traceback.format_exc())
        finally:
            # remove temporary logging file handler
            self.tcex.logger.remove_handler_by_name(self.thread_name)

            # unregister thread from token
            self.tcex.token.unregister_thread(trigger_id, self.thread_name)

    def fire_event_publish(self, trigger_id, session_id, request_key=None):
        """Send FireEvent command.

        Args:
            trigger_id (int): The ID of the trigger.
            session_id (str): The generated session for this fired event.
            request_key (str): The request key for this response.
        """
        msg = {
            'command': 'FireEvent',
            'triggerId': trigger_id,  # reference to single playbook
            'sessionId': session_id,  # session for the playbook execution
        }
        if request_key is not None:
            msg['requestKey'] = request_key  # reference for a specific playbook execution
        self.tcex.log.info(f'Firing Event ({msg})')

        # publish FireEvent command to client topic
        self.publish(json.dumps(msg))

    def format_query_string(self, params):
        """Convert name/value array to a query string.

        Args:
            params (dict): The query params for the request.

        Returns:
            str: The query params reformatted as a string.
        """
        query_string = []
        try:
            for q in params:
                query_string.append(f"{q.get('name')}={q.get('value')}")
        except AttributeError as e:
            self.tcex.log.error(f'Bad params data provided {params} ({e})')
            self.tcex.log.trace(traceback.format_exc())
        return '&'.join(query_string)

    def format_request_headers(self, headers):
        """Convert name/value array to a headers dict.

        Args:
            headers (dict): The dict of key/value header data.

        Returns:
            dict: The restructured header data.
        """
        headers_ = {}
        try:
            for h in headers:
                # TODO: either support tuple or csv list of values
                # headers_.setdefault(h.get('name').lower(), []).append(h.get('value'))
                headers_.setdefault(h.get('name').lower(), str(h.get('value')))

        except AttributeError as e:
            self.tcex.log.error(f'Bad header data provided {headers} ({e})')
            self.tcex.log.trace(traceback.format_exc())
        return headers_

    def format_response_headers(self, headers):
        """Convert name/value array to a query string.

        Args:
            headers (dict): The dict header data to be converted to key/value pairs.

        Returns:
            dict: The restructured header data.
        """
        headers_ = []
        try:
            for h in headers:
                headers_.append({'name': h[0], 'value': h[1]})
        except AttributeError as e:
            self.tcex.log.error(f'Bad header data provided {headers} ({e})')
            self.tcex.log.trace(traceback.format_exc())
        return headers_

    def heartbeat(self):
        """Start heartbeat process."""
        # start heartbeat monitor thread
        t = threading.Thread(name='heartbeat', target=self.heartbeat_monitor)
        t.daemon = True  # use setter for py2
        t.start()
        self.tcex.log.info('Heartbeat monitor started')

    def heartbeat_monitor(self):
        """Publish heartbeat on timer."""
        while True:
            if self.heartbeat_watchdog > (
                int(self.tcex.default_args.tc_svc_hb_timeout_seconds)
                / int(self.heartbeat_sleep_time)
            ):
                self.tcex.log.error('Missed server heartbeat message. Service is shutting down.')
                self.process_shutdown('Missed heartbeat commands.')
                break
            time.sleep(self.heartbeat_sleep_time)
            self.heartbeat_watchdog += 1

    def increment_metric(self, label, value=1):
        """Increment a metric if already exists.

        Args:
            label (str): The metric label (e.g., hits) to increment.
            value (int): The increment value. Defaults to 1.
        """
        if self._metrics.get(label) is not None:
            self._metrics[label] += value

    def listen(self):
        """List for message coming from broker."""
        self.tcex.log.trace(f'listen with {self.tcex.args.tc_svc_broker_service} broker')
        if self.tcex.args.tc_svc_broker_service.lower() == 'mqtt':
            target = self.listen_mqtt
        elif self.tcex.args.tc_svc_broker_service.lower() == 'redis':
            target = self.listen_redis

        t = threading.Thread(name='broker-listener', target=target, args=())
        t.daemon = True  # use setter for py2
        t.start()

    def listen_mqtt(self):
        """Listen for message coming from broker."""
        try:
            self.mqtt_client.on_connect = self.on_connect
            self.mqtt_client.on_message = self.on_message_mqtt

            # only needed when debugging
            # if self.tcex.log.getEffectiveLevel() == 5:
            #     self.mqtt_client.on_disconnect = self.on_disconnect
            #     self.mqtt_client.on_log = self.on_log
            #     self.mqtt_client.on_publish = self.on_publish
            #     self.mqtt_client.on_subscribe = self.on_subscribe
            #     self.mqtt_client.on_unsubscribe = self.on_unsubscribe

            # handle connection issues by not using loop_forever. give the service X seconds to
            # connect to message broker, else timeout and log generic connection error.
            self.mqtt_client.loop_start()
            deadline = time.time() + self.tcex.args.tc_svc_broker_conn_timeout
            while True:
                if not self._connected and deadline < time.time():
                    self.mqtt_client.loop_stop()
                    raise ConnectionError(
                        f'failed to connect to message broker host '
                        f'{self.tcex.args.tc_svc_broker_host} on port '
                        f'{self.tcex.args.tc_svc_broker_port}.'
                    )
                time.sleep(1)

        except Exception as e:
            self.tcex.log.trace(f'error in listen_mqtt: {e}')
            self.tcex.log.error(traceback.format_exc())

    def listen_redis(self):
        """Listen for message coming from broker."""
        self.tcex.log.info(f'Listening Redis topic {self.tcex.default_args.tc_svc_server_topic}')
        p = self.redis_client.pubsub(ignore_subscribe_messages=True)
        p.subscribe(self.tcex.default_args.tc_svc_server_topic)
        for message in p.listen():
            self.on_message_redis(message)

    def loop_forever(self, sleep=1):
        """Block and wait for shutdown.

        Args:
            sleep (int, optional): The amount of time to sleep between iterations. Defaults to 1.

        Returns:
            Bool: Returns True until shutdown received.
        """
        while True:
            deadline = time.time() + sleep
            while time.time() < deadline:
                if self.shutdown:
                    return False
                time.sleep(1)
            return True

    def message_thread(self, name, target, args, kwargs=None):
        """Start a message thread.

        Args:
            name (str): The name of the thread.
            target (callable): The method to call for the thread.
            args (tuple): The args to pass to the target method.
            kwargs (dict): Additional args.
        """
        # self.tcex.log.trace(f'message thread: {type(target)} - {args}')
        try:
            t = threading.Thread(name=name, target=target, args=args, kwargs=kwargs)
            t.daemon = True  # use setter for py2
            t.start()
        except Exception:
            self.tcex.log.trace(traceback.format_exc())

    @property
    def metrics(self):
        """Return current metrics."""
        if self._metrics.get('Active Playbooks') is not None:
            self.update_metric('Active Playbooks', len(self.configs))
        return self._metrics

    @metrics.setter
    def metrics(self, metrics):
        """Return current metrics."""
        if isinstance(metrics, dict):
            self._metrics = metrics
        else:
            self.tcex.log.error('Invalid data provided for metrics.')

    @property
    def mqtt_client(self):
        """Return a configured instance of mqtt client."""
        if self._mqtt_client is None:
            try:
                self._mqtt_client = mqtt.Client(client_id='', clean_session=True)
                self._mqtt_client.connect(
                    self.tcex.args.tc_svc_broker_host,
                    int(self.tcex.args.tc_svc_broker_port),
                    int(self.tcex.args.tc_svc_broker_timeout),
                )
                self._mqtt_client.tls_set(
                    ca_certs=self.tcex.args.tc_svc_broker_cacert_file,
                    cert_reqs=ssl.CERT_REQUIRED,
                    tls_version=ssl.PROTOCOL_TLSv1_2,
                )
                # add logger
                self._mqtt_client.enable_logger(logger=self.tcex.log)
                # username must be a empty string
                self._mqtt_client.username_pw_set('', password=self.tcex.args.tc_svc_broker_token)
                self._mqtt_client.tls_insecure_set(False)
            except Exception as e:
                self.tcex.log.error(f'Failed connection to MQTT service ({e})')
                self.tcex.log.trace(traceback.format_exc())
                self.process_shutdown('Failure connecting to mqtt')

        return self._mqtt_client

    def on_connect(self, client, userdata, flags, rc):  # pylint: disable=unused-argument
        """On connect method for mqtt broker."""
        self.tcex.log.info(f'Message broker connection status: {str(rc)}')

        self.tcex.log.info(f'Subscribing to topic: {self.tcex.default_args.tc_svc_server_topic}')
        self.mqtt_client.subscribe(self.tcex.default_args.tc_svc_server_topic)
        self._connected = True
        self.mqtt_client.disable_logger()

    def on_log(self, client, userdata, level, buf):  # pylint: disable=unused-argument
        """Handle MQTT on_log events."""
        self.tcex.log.trace(f'on_log - buf: {buf}, level: {level}')

    def on_message_mqtt(self, client, userdata, message):  # pylint: disable=unused-argument
        """On message for mqtt."""
        self.tcex.log.trace(f'on_message - message.payload: {message.payload}')
        self.tcex.log.trace(f'on_message - message.topic: {message.topic}')
        try:
            # messages on server topic must be json objects
            m = json.loads(message.payload)
        except ValueError:
            self.tcex.log.warning(f'Cannot parse message ({m}).')
            return

        if message.topic == self.tcex.default_args.tc_svc_server_topic:
            self.server_topic(m)

    def on_message_redis(self, message):  # pylint: disable=unused-argument
        """Subscribe and listen to "message" on Redis topic."""
        self.tcex.log.trace(f'on_message - message {message}')
        # only process "message" on topic (exclude subscriptions, etc)
        if message.get('type') != 'message':
            return

        try:
            # load message data
            m = json.loads(message.get('data'))
        except ValueError:
            self.tcex.log.warning(f'Cannot parse message ({message}).')
            return
        self.server_topic(m)

    def on_publish(self, client, userdata, result):  # pylint: disable=unused-argument
        """Handle MQTT on_log events."""
        self.tcex.log.trace(f'on_publish - {result}')

    def on_subscribe(self, client, userdata, mid, granted_qos):  # pylint: disable=unused-argument
        """Handle MQTT on_log events."""
        self.tcex.log.trace(f'on_subscribe - mid: {mid}, granted_qos: {granted_qos}')

    def process_config(self, message):
        """Process config message.

        Args:
            message (dict): The broker message.
        """
        command = message.get('command')
        config = message.get('config')
        status = 'Success'
        trigger_id = message.get('triggerId')

        if command.lower() == 'createconfig':
            self.tcex.log.info(f'CreateConfig - trigger_id: {trigger_id} config : {config}')
            valid_config = True

            # register config apiToken
            self.tcex.token.register_token(
                trigger_id, message.get('apiToken'), message.get('expireSeconds')
            )

            if callable(self.create_config_callback):
                msg = 'Config created'
                kwargs = {}
                if self.tcex.ij.runtime_level.lower() == 'webhooktriggerservice':
                    kwargs['url'] = message.get('url')
                try:
                    # call callback for create config and handle exceptions to protect thread
                    valid_config = self.create_config_callback(  # pylint: disable=not-callable
                        trigger_id, config, **kwargs
                    )
                except Exception as e:
                    self.tcex.log.error(
                        f'The create config callback method encountered an error ({e}).'
                    )
                    self.tcex.log.trace(traceback.format_exc())
                    status = 'Failed'
                    valid_config = False

            # create config after callback to report status and message
            if valid_config is not False:
                self.create_config(trigger_id, config, msg, status)
        elif command.lower() == 'deleteconfig':
            self.tcex.log.info(f'DeleteConfig - trigger_id: {trigger_id}')

            # unregister config apiToken
            self.tcex.token.unregister_token(trigger_id)

            if callable(self.delete_config_callback):
                message = 'Delete created'
                try:
                    # call callback for delete config and handle exceptions to protect thread
                    self.delete_config_callback(trigger_id)  # pylint: disable=not-callable
                except Exception as e:
                    message = f'The delete config callback method encountered an error ({e}).'
                    self.tcex.log.error(message)
                    self.tcex.log.trace(traceback.format_exc())
                    status = 'Failed'

            # delete config
            self.delete_config(trigger_id, message, status)

    def process_run_service(self, message):
        """Process Webhook event messages.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            {
              "command": "RunService",
              "apiToken": "abc123",
              "bodyVariable": "request.body",
              "headers": [ { key/value pairs } ],
              "method": "GET",
              "queryParams": [ { key/value pairs } ],
              "requestKey": "123abc"
            }

        Args:
            message (dict): The broker message.
        """
        # register config apiToken (before any logging)
        self.tcex.token.register_token(
            self.thread_name, message.get('apiToken'), message.get('expireSeconds')
        )
        self.tcex.log.info('Processing RunService Command')
        self.tcex.log.debug(f'message: {message}')

        # thread event used to block response until body is written
        e = threading.Event()

        # process message
        request_key = message.get('requestKey')
        body = None
        try:
            # read body from redis
            body_variable = message.get('bodyVariable')
            if body_variable is not None:
                body = self.redis_client.hget(request_key, message.get('bodyVariable'))
                if body is not None:
                    body = StringIO(base64.b64decode(body).decode('utf-8'))
        except Exception as e:
            self.tcex.log.error(f'Failed reading body to Redis ({e})')
            self.tcex.log.trace(traceback.format_exc())
        headers = self.format_request_headers(message.get('headers'))
        method = message.get('method')
        params = message.get('queryParams')
        path = message.get('path')

        try:
            # TODO: research required field for wsgi and update
            # TODO: move to a method
            environ = {
                'wsgi.errors': self.tcex.log.error,  # sys.stderr
                # 'wsgi.file_wrapper': <class 'wsgiref.util.FileWrapper'>
                'wsgi.input': body,
                'wsgi.multithread': True,
                'wsgi.multiprocess': False,
                'wsgi.run_once': True,
                'wsgi.url_scheme': 'https',
                'wsgi.version': (1, 0),
                # 'GATEWAY_INTERFACE': 'CGI/1.1',
                # 'HTTP_ACCEPT': '',
                'HTTP_ACCEPT': headers.get('accept', ''),
                # 'HTTP_ACCEPT_ENCODING': '',
                # 'HTTP_ACCEPT_LANGUAGE': '',
                # 'HTTP_COOKIE': '',
                # 'HTTP_DNT': 1,
                # 'HTTP_CONNECTION': 'keep-alive',
                # 'HTTP_HOST': '',
                # 'HTTP_UPGRADE_INSECURE_REQUESTS': 1,
                'HTTP_USER_AGENT': headers.get('user-agent', ''),
                'PATH_INFO': path,
                'QUERY_STRING': self.format_query_string(params),
                # 'REMOTE_ADDR': '',
                # 'REMOTE_HOST': '',
                'REQUEST_METHOD': method.upper(),
                'SCRIPT_NAME': '/',
                'SERVER_NAME': '',
                'SERVER_PORT': '',
                'SERVER_PROTOCOL': 'HTTP/1.1',
                # 'SERVER_SOFTWARE': 'WSGIServer/0.2',
            }
            if headers.get('content-type') is not None:
                environ['CONTENT_TYPE'] = (headers.get('content-type'),)
            if headers.get('content-length') is not None:
                environ['CONTENT_LENGTH'] = headers.get('content-length')
            self.tcex.log.trace(f'environ: {environ}')
            self.increment_metric('Requests')
        except Exception as e:
            self.tcex.log.error(f'Failed building environ ({e})')
            self.tcex.log.trace(traceback.format_exc())
            self.increment_metric('Errors')
            return  # stop processing

        def response_handler(*args, **kwargs):  # pylint: disable=unused-argument
            """Handle WSGI Response"""
            kwargs['e'] = e  # add event to kwargs for blocking
            kwargs['request_key'] = request_key
            t = threading.Thread(
                name='response-handler',
                target=self.process_run_service_response,
                args=args,
                kwargs=kwargs,
            )
            t.daemon = True  # use setter for py2
            t.start()

        if callable(self.api_event_callback):
            try:
                body = self.api_event_callback(  # pylint: disable=not-callable
                    environ, response_handler
                )

                # decode body entries
                # TODO: validate this logic
                body = [base64.b64encode(b).decode('utf-8') for b in body][0]
                # write body to Redis
                self.redis_client.hset(request_key, 'response.body', body)

                # set thread event to True to trigger response
                self.tcex.log.info('API response body written')
                e.set()
            except Exception as e:
                self.tcex.log.error(f'The api event callback method encountered and error ({e}).')
                self.tcex.log.trace(traceback.format_exc())
                self.increment_metric('Errors')

        # unregister config apiToken
        self.tcex.token.unregister_token(self.thread_name)

    def process_run_service_response(self, *args, **kwargs):
        """Handle service event responses.

        ('200 OK', [('content-type', 'application/json'), ('content-length', '103')])
        """
        self.tcex.log.info('API response received, waiting on body to be written')
        kwargs.get('e').wait(10)  # wait for thread event - (set on body write)
        self.tcex.log.trace(f'response args: {args}')
        try:
            status_code, status = args[0].split(' ', 1)
            response = {
                'bodyVariable': 'response.body',
                'command': 'Acknowledged',
                'headers': self.format_response_headers(args[1]),
                'requestKey': kwargs.get('request_key'),  # pylint: disable=cell-var-from-loop
                'status': status,
                'statusCode': status_code,
                'type': 'RunService',
            }
            self.tcex.log.info('API response sent')
            self.publish(json.dumps(response))
            self.increment_metric('Responses')
        except Exception as e:
            self.tcex.log.error(f'Failed creating response body ({e})')
            self.tcex.log.trace(traceback.format_exc())
            self.increment_metric('Errors')

    def process_shutdown(self, reason):
        """Handle a shutdown message.

        Args:
            reason (str): The reason for the shutdown.
        """
        reason = reason or (
            'A shutdown command was received on server topic. Service is shutting down.'
        )
        self.tcex.log.info(f'Shutdown - reason: {reason}')

        # acknowledge shutdown command
        self.publish(json.dumps({'command': 'Acknowledged', 'type': 'Shutdown'}))

        # call App shutdown callback
        if callable(self.shutdown_callback):
            try:
                # call callback for shutdown and handle exceptions to protect thread
                self.shutdown_callback()  # pylint: disable=not-callable
            except Exception as e:
                self.tcex.log.error(f'The shutdown callback method encountered and error ({e}).')
                self.tcex.log.trace(traceback.format_exc())

        # unsubscribe and disconnect from the broker
        if self.tcex.args.tc_svc_broker_service.lower() == 'mqtt':
            self.mqtt_client.unsubscribe(self.tcex.default_args.tc_svc_server_topic)
            self.mqtt_client.disconnect()

        # update shutdown flag
        self.shutdown = True

        # delay shutdown to give App time to cleanup
        time.sleep(5)
        self.tcex.exit(0)  # final shutdown in case App did not

    def process_webhook(self, message):
        """Process Webhook event messages.

        Args:
            message (dict): The message data from the broker.
        """
        # session auth shared data is thread name which needs to map back to config triggerId
        self.tcex.token.register_thread(message.get('triggerId'), self.thread_name)

        # add logger for current session
        self.tcex.logger.add_thread_file_handler(
            name=self.thread_name,
            filename=self.session_logfile,
            level=self.tcex.default_args.tc_log_level,
            path=self.tcex.default_args.tc_log_path,
        )
        self.tcex.log.trace('Process webhook event trigger')

        # acknowledge webhook event
        self.publish(
            json.dumps(
                {
                    'command': 'Acknowledged',
                    'requestKey': message.get('requestKey'),
                    'triggerId': message.get('triggerId'),
                    'type': 'WebhookEvent',
                }
            )
        )

        # get config using triggerId passed in WebhookEvent data
        config = self.configs.get(message.get('triggerId'))
        if config is None:
            self.tcex.log.error(f"Could not find config for triggerId {message.get('triggerId')}")

        # get an instance of playbooks for App
        outputs = config.get('tc_playbook_out_variables') or []
        if isinstance(outputs, str):
            outputs = outputs.split(',')

        playbook = self.tcex.pb(context=self.thread_name, output_variables=outputs)

        try:
            request_key = message.get('requestKey')
            body = self.redis_client.hget(request_key, 'request.body')
            if body is not None:
                body = base64.b64decode(body).decode()
            headers = message.get('headers')
            method = message.get('method')
            params = message.get('queryParams')
            trigger_id = message.get('triggerId')
            callback_response = self.webhook_event_callback(  # pylint: disable=not-callable
                trigger_id, playbook, method, headers, params, body, config
            )
            if isinstance(callback_response, dict):
                # webhook responses are for providers that require a subscription req/resp.
                webhook_event_response = {
                    'sessionId': self.thread_name,  # session/context
                    'requestKey': request_key,
                    'command': 'WebhookEventResponse',
                    'triggerId': trigger_id,
                    'bodyVariable': 'response.body',
                    'headers': callback_response.get('headers', []),
                    'statusCode': callback_response.get('statusCode', 200),
                }
                # write response body to redis
                if callback_response.get('body') is not None:
                    playbook.create_string(
                        'response.body',
                        base64.b64encode(callback_response.get('body').encode('utf-8')).decode(
                            'utf-8'
                        ),
                    )

                # publish the WebhookEventResponse message
                self.publish(json.dumps(webhook_event_response))
            elif isinstance(callback_response, bool) and callback_response:
                self.increment_metric('Hits')
                self.fire_event_publish(trigger_id, self.thread_name, request_key)

                # only required for testing in tcex framework
                self._tcex_testing(self.thread_name, trigger_id)
            else:
                self.increment_metric('Misses')
        except Exception as e:
            self.tcex.log.error(f'The callback method encountered and error ({e}).')
            self.tcex.log.trace(traceback.format_exc())
            self.increment_metric('Errors')
        finally:
            # remove temporary logging file handler
            self.tcex.logger.remove_handler_by_name(self.thread_name)

            # unregister thread from token
            self.tcex.token.unregister_thread(message.get('triggerId'), self.thread_name)

    def publish(self, message, topic=None):
        """Publish a message on client topic.

        Args:
            message (str): The message to be sent on client topic.
            topic (str): The broker topic. Default to None.
        """
        if topic is None:
            topic = self.tcex.default_args.tc_svc_client_topic
        self.tcex.log.debug(f'publish topic: ({topic})')
        self.tcex.log.debug(f'publish message: ({message})')

        if self.tcex.args.tc_svc_broker_service.lower() == 'mqtt':
            r = self.mqtt_client.publish(topic, message)
            self.tcex.log.trace(f'publish response: {r}')
        elif self.tcex.args.tc_svc_broker_service.lower() == 'redis':
            self.redis_client.publish(topic, message)

    @property
    def ready(self):
        """Return ready boolean."""
        return self._ready

    @ready.setter
    def ready(self, bool_val):
        """Set ready boolean."""
        if isinstance(bool_val, bool) and bool_val is True:
            # wait until connected to send ready command
            while not self._connected:
                if self.shutdown:
                    break
                time.sleep(1)
            else:  # pylint: disable=useless-else-on-loop
                self.tcex.log.info('Service is Ready')
                self.publish(json.dumps({'command': 'Ready'}))
                self._ready = True

    @property
    def redis_client(self):
        """Return the correct KV store for this execution."""
        if self._redis_client is None:
            self._redis_client = self.tcex.redis_client
        return self._redis_client

    def server_topic(self, message):
        """Handle any event coming in on server_topic.

        Args:
            message (dict): The broker message.
        """
        self.tcex.log.trace(f'message: {message}')
        # parse the command type
        command = message.get('command')

        self.tcex.log.info(f'Command received: {command}')
        if command.lower() == 'heartbeat':
            self.heartbeat_watchdog = 0
            self.heartbeat_miss_count = 0

            # send heartbeat -acknowledge- command
            response = {'command': 'Heartbeat', 'metric': self.metrics}
            self.publish(json.dumps(response))
            self.tcex.log.info('Heartbeat command sent')
            self.tcex.log.debug(f'metrics: {self.metrics}')
        elif command.lower() == 'loggingchange':
            # {"command": "LoggingChange", "level": "DEBUG"}
            level = message.get('level')
            self.tcex.log.info(f'LoggingChange - level: {level}')
            self.tcex.logger.update_handler_level(level)
        elif command.lower() == 'runservice':
            self.message_thread(
                self.session_id(message.get('triggerId')), self.process_run_service, (message,)
            )
        elif command.lower() == 'shutdown':
            # {"command": "Shutdown", "reason": "Service disabled by user."}
            reason = message.get('reason')
            self.process_shutdown(reason)
        elif command.lower() == 'webhookevent':
            self.message_thread(
                self.session_id(message.get('triggerId')), self.process_webhook, (message,)
            )
        else:
            # any other message is a config message
            self.message_thread('process-config', self.process_config, (message,))

    @staticmethod
    def session_id(trigger_id=None):  # pylint: disable=unused-argument
        """Return a uuid4 session id.

        Args:
            trigger_id (str): Optional trigger_id value used in testing framework.
        """
        return str(uuid.uuid4())

    @property
    def session_logfile(self):
        """Return a uuid4 session id."""
        return f"{datetime.today().strftime('%Y%m%d')}/{self.thread_name}.log"

    @property
    def thread_name(self):
        """Return a uuid4 session id."""
        return threading.current_thread().name

    def update_metric(self, label, value):
        """Update a metric if already exists.

        Args:
            label (str): The metric label (e.g., hits) to update.
            value (int|str): The updated value for the metric.
        """
        if self._metrics.get(label) is not None:
            self._metrics[label] = value



# -*- coding: utf-8 -*-
"""TcEx Service Common Module"""
import base64
import json
import os
import subprocess
import sys
import threading
import time
import uuid
from multiprocessing import Process
from random import randint

import paho.mqtt.client as mqtt

from .test_case_playbook_common import TestCasePlaybookCommon


class TestCaseServiceCommon(TestCasePlaybookCommon):
    """Service App TestCase Class"""

    _mqtt_client = None
    app_process = None
    client_topic = f'client-topic-{randint(100, 999)}'
    server_topic = f'server-topic-{randint(100, 999)}'
    service_run_method = 'subprocess'  # run service as subprocess, multiprocess, or thread
    shutdown = False
    shutdown_complete = False
    sleep_after_publish_config = 0.5
    sleep_after_publish_webhook_event = 0.5
    sleep_after_service_start = 5
    sleep_before_delete_config = 2
    sleep_before_shutdown = 0.5

    def _app_callback(self, app):
        """Set app object from run.py callback"""
        self.app = app

    @property
    def default_args(self):
        """Return App default args."""
        args = super().default_args.copy()
        args.update(
            {
                'tc_svc_broker_host': os.getenv('TC_SVC_BROKER_HOST', 'localhost'),
                'tc_svc_broker_port': int(os.getenv('TC_SVC_BROKER_PORT', '1883')),
                'tc_svc_broker_service': os.getenv('TC_SVC_BROKER_SERVICE', 'mqtt'),
                'tc_svc_broker_token': os.getenv('TC_SVC_BROKER_TOKEN'),
                'tc_svc_client_topic': self.client_topic,
                'tc_svc_server_topic': self.server_topic,
                'tc_svc_hb_timeout_seconds': int(os.getenv('TC_SVC_HB_TIMEOUT_SECONDS', '300')),
            }
        )
        return args

    @property
    def mqtt_client(self):
        """Return a mqtt client instance."""
        if self._mqtt_client is None:
            self._mqtt_client = mqtt.Client()
            self._mqtt_client.connect(
                self.tcex.args.tc_svc_broker_host,
                self.tcex.args.tc_svc_broker_port,
                self.tcex.args.tc_svc_broker_timeout,
            )
        return self._mqtt_client

    def publish(self, message, topic=None):
        """Publish message on server channel.

        Args:
            message (str): The message to send.
            topic (str, optional): The message broker topic. Defaults to None.
        """
        topic = topic or self.server_topic

        # self.log.debug(f'topic: ({topic})')
        # self.log.debug(f'message: ({message})')
        # self.log.debug(f'broker_service: ({self.tcex.args.tc_svc_broker_service})')

        if self.tcex.args.tc_svc_broker_service.lower() == 'mqtt':
            self.mqtt_client.publish(topic, message)
        elif self.tcex.args.tc_svc_broker_service.lower() == 'redis':
            self.redis_client.publish(topic, message)

    def publish_create_config(self, message):
        """Send create config message.

        Args:
            trigger_id (str): The trigger id for the config message.
            message (dict): The entire message with trigger_id and config.
        """
        # merge the message config (e.g., optional, required)
        message_config = message.pop('config')
        config = message_config.get('optional', {})
        config.update(message_config.get('required', {}))
        message['config'] = config

        # build config message
        message['apiToken'] = self.tc_token
        message['expireSeconds'] = int(time.time() + 86400)
        message['command'] = 'CreateConfig'
        message['config']['tc_playbook_out_variables'] = self.profile.tc_playbook_out_variables
        message['triggerId'] = message.pop('trigger_id')
        self.publish(json.dumps(message))
        time.sleep(self.sleep_after_publish_config)

    def publish_delete_config(self, message):
        """Send delete config message.

        Args:
            message (str): The message coming in on Broker channel
        """
        time.sleep(self.sleep_before_delete_config)
        # using triggerId here instead of trigger_id do to pop in publish_create_config
        config_msg = {'command': 'DeleteConfig', 'triggerId': message.get('triggerId')}
        self.publish(json.dumps(config_msg))

    def publish_shutdown(self):
        """Publish shutdown message."""
        config_msg = {'command': 'Shutdown'}
        self.publish(json.dumps(config_msg))

    def publish_webhook_event(
        self,
        trigger_id,
        body=None,
        headers=None,
        method='GET',
        query_params=None,
        request_key='abc123',
    ):
        """Send create config message.

        Args:
            trigger_id (str): The trigger ID.
            body (str): The Body of the request.
            headers (list, optional): A list of headers name/value pairs. Defaults to [].
            method (str, optional): The method. Defaults to 'GET'.
            query_params (list, optional): A list of query param name/value pairs. Defaults to [].
            request_key (str, optional): The current request key.
        """
        body = body or ''
        if isinstance(body, dict):
            body = json.dumps(body)

        body = self.redis_client.hset(
            request_key, 'request.body', base64.b64encode(body.encode('utf-8'))
        )
        event = {
            'command': 'WebhookEvent',
            'method': method,
            'queryParams': query_params or [],
            'headers': headers or [],
            'body': 'request.body',
            'requestKey': request_key,
            'triggerId': trigger_id,
        }
        self.publish(json.dumps(event))
        time.sleep(self.sleep_after_publish_webhook_event)

    def run(self):
        """Implement in Child Class"""
        raise NotImplementedError('Child class must implement this method.')

    def run_service(self):
        """Run the micro-service."""
        self.log.data('run', 'service method', self.service_run_method)
        # backup sys.argv
        sys_argv_orig = sys.argv

        # clear sys.argv
        sys.argv = sys.argv[:1]

        # create required .app_params encrypted file. args are set in custom.py
        self.args['tcex_testing_context'] = self.tcex_testing_context
        self.create_config(self.args)

        # run the service App in 1 of 3 ways
        if self.service_run_method == 'subprocess':
            # run the Service App as a subprocess
            self.app_process = subprocess.Popen(['python', 'run.py'])
        elif self.service_run_method == 'thread':

            # run App in a thread
            t = threading.Thread(target=self.run, args=(), daemon=True)
            t.start()
        elif self.service_run_method == 'multiprocess':
            p = Process(target=self.run, args=(), daemon=True)
            p.start()

        # give app some time to initialize before continuing
        time.sleep(self.sleep_after_service_start)

        # restore sys.argv
        sys.argv = sys_argv_orig

    @classmethod
    def setup_class(cls):
        """Run once before all test cases."""
        super().setup_class()
        cls.args = {}
        cls.service_file = 'SERVICE_STARTED'  # started file flag

        # generate a context used in service.py to write context during fire event
        cls.tcex_testing_context = str(uuid.uuid4())

    def setup_method(self):
        """Run before each test method runs."""
        super().setup_method()
        self.stager.redis.from_dict(self.redis_staging_data)
        self.redis_client = self.tcex.redis_client

        # only start service if it hasn't been started already base on file flag.
        if not os.path.isfile(self.service_file):
            open(self.service_file, 'w+').close()  # create service started file flag
            self.run_service()

            # start shutdown monitor thread
            t = threading.Thread(target=self.shutdown_monitor, daemon=True)
            t.start()

    def shutdown_monitor(self):
        """Monitor for shutdown flag."""
        while not self.shutdown:
            time.sleep(0.5)

        # shutdown the App
        self.publish_shutdown()

        # give Service App x seconds to shutdown before terminating
        if self.service_run_method == 'subprocess':
            for _ in range(1, 10):
                time.sleep(0.5)
                if self.app_process.poll() is not None:
                    break
            else:
                self.log.data('run', 'Terminating Process', f'PID: {self.app_process.pid}', 'debug')
                self.app_process.terminate()  # terminate subprocess

        # remove started file flag
        try:
            os.remove(self.service_file)
        except OSError:
            pass

        # set shutdown_complete
        self.shutdown_complete = True

    def stage_data(self, staged_data):
        """Stage the data in the profile."""
        for key, value in list(staged_data.get('redis', {}).items()):
            self.stager.redis.stage(key, value)

    @classmethod
    def teardown_class(cls):
        """Run once before all test cases."""
        super().teardown_class()
        # set shutdown flag for shutdown_monitor and wait until shutdown is done
        cls.shutdown = True
        for _ in range(1, 12):
            if cls.shutdown_complete:
                break
            time.sleep(0.5)

    def teardown_method(self):
        """Run after each test method runs."""
        time.sleep(self.sleep_before_shutdown)
        # run test_case_playbook_common teardown_method
        super().teardown_method()

        # clean up tcex testing context after populate_output has run
        self.clear_context(self.tcex_testing_context)



# -*- coding: utf-8 -*-
"""API Handler Class"""
import logging
import os
import threading


class ThreadFileHandler(logging.FileHandler):
    """Logger handler for ThreatConnect Exchange File logging."""

    def __init__(self, filename, mode='a', encoding=None, delay=0):
        """Add logic to create log directory if it does not exists.

        Args:
            filename (str): The name of the logfile.
            mode (str, optional): The write mode for the file. Defaults to 'a'.
            encoding (str, optional): The log file encoding. Defaults to None.
            delay (int, optional): The delay period. Defaults to 0.
        """
        if not os.path.exists(os.path.dirname(filename)):  # pragma: no cover
            try:
                # pylint: disable=unexpected-keyword-arg
                os.makedirs(os.path.dirname(filename), exist_ok=True)
            except TypeError:
                # TODO: [py2] - remove py2 specific code and pylint-disable
                if not os.path.exists(os.path.dirname(filename)):
                    os.makedirs(os.path.dirname(filename))

        logging.FileHandler.__init__(self, filename, mode, encoding, delay)

    def emit(self, record):
        """Emit a record.

        Args:
            record (obj): The record to be logged.
        """
        if self.get_name() == threading.current_thread().name:
            logging.FileHandler.emit(self, record)
